# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import datetime
import logging
import time
import struct
from enum import Enum, IntEnum
from spinn_utilities.config_holder import get_config_bool
from spinn_utilities.overrides import overrides
from spinn_utilities.log import FormatAdapter
from spinnman.exceptions import SpinnmanTimeoutException
from spinnman.messages.sdp import SDPMessage, SDPHeader, SDPFlag
from spinnman.model.enums.cpu_state import CPUState
from pacman.model.graphs.machine import MachineVertex
from pacman.model.resources import ConstantSDRAM, IPtagResource
from spinn_front_end_common.data import FecDataView
from spinn_front_end_common.interface.provenance import ProvenanceWriter
from spinn_front_end_common.utilities.helpful_functions import (
    convert_vertices_to_core_subset, n_word_struct)
from spinn_front_end_common.utilities.emergency_recovery import (
    emergency_recover_state_from_failure)
from spinn_front_end_common.abstract_models import (
    AbstractHasAssociatedBinary, AbstractGeneratesDataSpecification)
from spinn_front_end_common.interface.provenance import (
    AbstractProvidesProvenanceDataFromMachine)
from spinn_front_end_common.utilities.utility_objs import (
    ExecutableType)
from spinn_front_end_common.utilities.constants import (
    SDP_PORTS, BYTES_PER_WORD, BYTES_PER_KB)
from spinn_front_end_common.utilities.utility_calls import (
    get_region_base_address_offset, open_scp_connection, retarget_tag)
from spinn_front_end_common.utilities.exceptions import SpinnFrontEndException
from spinn_front_end_common.utilities.utility_objs.\
    extra_monitor_scp_processes import (
        SetRouterTimeoutProcess, ClearQueueProcess)

log = FormatAdapter(logging.getLogger(__name__))

# shift by for the destination x coord in the word.
DEST_X_SHIFT = 16

TIMEOUT_RETRY_LIMIT = 100
_MINOR_LOSS_THRESHOLD = 10

# cap for stopping wrap arounds
TRANSACTION_ID_CAP = 0xFFFFFFFF

#: number of items used up by the retransmit code for its header
SDP_RETRANSMISSION_HEADER_SIZE = 2

#: size of config region in bytes
#: 1.new seq key, 2.first data key, 3. transaction id key 4.end flag key,
# 5.base key, 6.iptag tag
CONFIG_SIZE = 6 * BYTES_PER_WORD

#: items of data a SDP packet can hold when SCP header removed
WORDS_PER_FULL_PACKET = 68  # 272 bytes as removed SCP header

#: size of items the sequence number uses
SEQUENCE_NUMBER_SIZE_IN_ITEMS = 1

#: transaction id size in words
TRANSACTION_ID_SIZE_IN_ITEMS = 1

#: the size in words of the command flag
COMMAND_SIZE_IN_ITEMS = 1

#: offset for missing seq starts in first packet
WORDS_FOR_COMMAND_N_MISSING_TRANSACTION = 3

#: offset for missing seq starts in more packet
WORDS_FOR_COMMAND_TRANSACTION = (
    COMMAND_SIZE_IN_ITEMS + TRANSACTION_ID_SIZE_IN_ITEMS)

BYTES_FOR_SEQ_AND_TRANSACTION_ID = (
    (SEQUENCE_NUMBER_SIZE_IN_ITEMS + TRANSACTION_ID_SIZE_IN_ITEMS) *
    BYTES_PER_WORD)

#: items of data from SDP packet with a sequence number
WORDS_PER_FULL_PACKET_WITH_SEQUENCE_NUM = (
    WORDS_PER_FULL_PACKET - SEQUENCE_NUMBER_SIZE_IN_ITEMS -
    TRANSACTION_ID_SIZE_IN_ITEMS)

#: offset where data in starts on commands
#: (command, transaction_id, sequence number)
WORDS_FOR_COMMAND_AND_KEY = 3
BYTES_FOR_COMMAND_AND_ADDRESS_HEADER = (
    WORDS_FOR_COMMAND_AND_KEY * BYTES_PER_WORD)

#: offset where data in starts in reception (command, transaction id)
WORDS_FOR_RECEPTION_COMMAND_AND_ADDRESS_HEADER = 2
BYTES_FOR_RECEPTION_COMMAND_AND_ADDRESS_HEADER = (
    WORDS_FOR_RECEPTION_COMMAND_AND_ADDRESS_HEADER * BYTES_PER_WORD)

#: size for data to store when first packet with command and address
WORDS_IN_FULL_PACKET_WITH_KEY = (
    WORDS_PER_FULL_PACKET - WORDS_FOR_COMMAND_AND_KEY)
BYTES_IN_FULL_PACKET_WITH_KEY = (
    WORDS_IN_FULL_PACKET_WITH_KEY * BYTES_PER_WORD)

#: size of data in key space
#: x, y, key (all ints) for possible 48 chips, plus n chips to read,
# the reinjector base key.
SIZE_DATA_IN_CHIP_TO_KEY_SPACE = ((3 * 48) + 2) * BYTES_PER_WORD


class _DATA_REGIONS(IntEnum):
    """
    DSG data regions.
    """
    CONFIG = 0
    CHIP_TO_KEY_SPACE = 1
    PROVENANCE_REGION = 2


class DATA_OUT_COMMANDS(Enum):
    """
    Command IDs for the SDP packets for data out.
    """
    START_SENDING = 100
    START_MISSING_SEQ = 1000
    MISSING_SEQ = 1001
    CLEAR = 2000


class DATA_IN_COMMANDS(Enum):
    """
    Command IDs for the SDP packets for data in.
    """
    SEND_DATA_TO_LOCATION = 200
    SEND_SEQ_DATA = 2000
    SEND_TELL = 2001
    RECEIVE_MISSING_SEQ_DATA = 2002
    RECEIVE_FINISHED = 2003


# precompiled structures
_ONE_WORD = struct.Struct("<I")
_TWO_WORDS = struct.Struct("<II")
_THREE_WORDS = struct.Struct("<III")
_FOUR_WORDS = struct.Struct("<IIII")
_FIVE_WORDS = struct.Struct("<IIIII")


# Set to true to check that the data is correct after it has been sent in.
# This is expensive, and only works in Python 3.5 or later.
VERIFY_SENT_DATA = False

# provenance data size
_PROVENANCE_DATA_SIZE = 4 * BYTES_PER_WORD


def ceildiv(dividend, divisor):
    """
    How to divide two possibly-integer numbers and round up.
    """
    assert divisor > 0
    q, r = divmod(dividend, divisor)
    return int(q) + (r != 0)


# SDRAM requirement for storing missing SDP packets seq nums
SDRAM_FOR_MISSING_SDP_SEQ_NUMS = ceildiv(
    120.0 * 1024 * BYTES_PER_KB,
    WORDS_PER_FULL_PACKET_WITH_SEQUENCE_NUM * BYTES_PER_WORD)


class DataSpeedUpPacketGatherMachineVertex(
        MachineVertex, AbstractGeneratesDataSpecification,
        AbstractHasAssociatedBinary,
        AbstractProvidesProvenanceDataFromMachine):
    """
    Machine vertex for handling fast data transfer between host and  SpiNNaker.
    This machine vertex is only ever placed on chips with a working Ethernet
    connection; it collaborates with the
    :py:class:`ExtraMonitorSupportMachineVertex` to write data on other chips.

    .. note::
        This is an unusual machine vertex, in that it has no associated
        application vertex.
    """
    __slots__ = [
        # x coordinate
        "_x",
        # y coordinate
        "_y",
        # word with x and y
        "_coord_word",
        # transaction id
        "_transaction_id",
        # socket
        "_connection",
        # path for the data in report
        "_in_report_path",
        # ipaddress
        "_ip_address",
        # store for the last reinjection status
        "_last_status",
        # the max seq num expected given a data retrieval
        "_max_seq_num",
        # holder for missing seq nums for data in
        "_missing_seq_nums_data_in",
        # holder of data from out
        "_output",
        # my placement for future lookup
        "_placement",
        # Count of the runs for provenance data
        "_run",
        "_remote_tag",
        # path to the data out report
        "_out_report_path",
        # data holder for output
        "_view"]

    #: base key (really nasty hack to tie in fixed route keys)
    BASE_KEY = 0xFFFFFFF9
    NEW_SEQ_KEY = 0xFFFFFFF8
    FIRST_DATA_KEY = 0xFFFFFFF7
    END_FLAG_KEY = 0xFFFFFFF6
    TRANSACTION_ID_KEY = 0xFFFFFFF5

    #: to use with multicast stuff (reinjection acks have to be fixed route)
    BASE_MASK = 0xFFFFFFFB
    NEW_SEQ_KEY_OFFSET = 1
    FIRST_DATA_KEY_OFFSET = 2
    END_FLAG_KEY_OFFSET = 3
    TRANSACTION_ID_KEY_OFFSET = 4

    # throttle on the transmission
    _TRANSMISSION_THROTTLE_TIME = 0.000001

    #: report name for tracking used routers
    OUT_REPORT_NAME = "routers_used_in_speed_up_process.rpt"
    #: report name for tracking performance gains
    IN_REPORT_NAME = "speeds_gained_in_speed_up_process.rpt"

    # the end flag is set when the high bit of the sequence number word is set
    _LAST_MESSAGE_FLAG_BIT_MASK = 0x80000000
    # corresponding mask for the actual sequence numbers
    _SEQUENCE_NUMBER_MASK = 0x7fffffff

    # time outs used by the protocol for separate bits
    _TIMEOUT_PER_RECEIVE_IN_SECONDS = 2
    _TIMEOUT_FOR_SENDING_IN_SECONDS = 0.01

    # end flag for missing seq nums
    _MISSING_SEQ_NUMS_END_FLAG = 0xFFFFFFFF

    # flag for saying missing all SEQ numbers
    FLAG_FOR_MISSING_ALL_SEQUENCES = 0xFFFFFFFE

    _ADDRESS_PACKET_BYTE_FORMAT = struct.Struct(
        f"<{BYTES_IN_FULL_PACKET_WITH_KEY}B")

    # Router timeouts, in mantissa,exponent form. See datasheet for details
    _LONG_TIMEOUT = (14, 14)
    _SHORT_TIMEOUT = (1, 1)
    _TEMP_TIMEOUT = (15, 4)
    _ZERO_TIMEOUT = (0, 0)

    # Initial port for the reverse IP tag (to be replaced later)
    _TAG_INITIAL_PORT = 10000

    def __init__(self, x, y, ip_address):
        """
        :param int x: Where this gatherer is.
        :param int y: Where this gatherer is.
        :param str ip_address:
            How to talk directly to the chip where the gatherer is.
        """
        super().__init__(
            label=f"SYSTEM:PacketGatherer({x},{y})", app_vertex=None)

        # data holders for the output, and sequence numbers
        self._view = None
        self._max_seq_num = None
        self._output = None

        self._transaction_id = 0

        self._missing_seq_nums_data_in = list()

        # Create a connection to be used
        self._x = x
        self._y = y
        self._coord_word = None
        self._ip_address = ip_address
        self._remote_tag = None
        self._connection = None

        # local provenance storage
        self._run = 0
        self._placement = None

        # create report if it doesn't already exist

        dir_path = FecDataView.get_run_dir_path()
        self._out_report_path = os.path.join(dir_path, self.OUT_REPORT_NAME)
        self._in_report_path = os.path.join(dir_path, self.IN_REPORT_NAME)

        # Stored reinjection status for resetting timeouts
        self._last_status = None

    def __throttled_send(self, message):
        """
        Slows down transmissions to allow SpiNNaker to keep up.

        :param ~.SDPMessage message: message to send
        """
        # send first message
        self._connection.send_sdp_message(message)
        time.sleep(self._TRANSMISSION_THROTTLE_TIME)

    @property
    @overrides(MachineVertex.sdram_required)
    def sdram_required(self):
        return ConstantSDRAM(
                CONFIG_SIZE + SDRAM_FOR_MISSING_SDP_SEQ_NUMS +
                SIZE_DATA_IN_CHIP_TO_KEY_SPACE + _PROVENANCE_DATA_SIZE)

    @property
    @overrides(MachineVertex.iptags)
    def iptags(self):
        return [IPtagResource(
            port=self._TAG_INITIAL_PORT, strip_sdp=True,
            ip_address="localhost", traffic_identifier="DATA_SPEED_UP")]

    def update_transaction_id_from_machine(self):
        """
        Looks up from the machine what the current transaction ID is
        and updates the data speed up gatherer.
        """
        self._transaction_id = FecDataView.get_transceiver().read_user_1(
            self._placement.x, self._placement.y, self._placement.p)

    @overrides(AbstractHasAssociatedBinary.get_binary_start_type)
    def get_binary_start_type(self):
        return ExecutableType.SYSTEM

    @overrides(AbstractGeneratesDataSpecification.generate_data_specification)
    def generate_data_specification(self, spec, placement):
        # pylint: disable=unsubscriptable-object
        # update my placement for future knowledge
        self._placement = placement

        # Create the data regions for hello world
        self._reserve_memory_regions(spec)

        # the keys for the special cases
        new_seq_key = self.NEW_SEQ_KEY
        first_data_key = self.FIRST_DATA_KEY
        end_flag_key = self.END_FLAG_KEY
        base_key = self.BASE_KEY
        transaction_id_key = self.TRANSACTION_ID_KEY

        spec.switch_write_focus(_DATA_REGIONS.CONFIG)
        spec.write_value(new_seq_key)
        spec.write_value(first_data_key)
        spec.write_value(transaction_id_key)
        spec.write_value(end_flag_key)
        spec.write_value(base_key)

        # locate the tag ID for our data and update with a port
        # Note: The port doesn't matter as we are going to override this later
        iptags = FecDataView.get_tags().get_ip_tags_for_vertex(self)
        iptag = iptags[0]
        spec.write_value(iptag.tag)
        self._remote_tag = iptag.tag

        # write mc chip key map
        machine = FecDataView.get_machine()
        spec.switch_write_focus(_DATA_REGIONS.CHIP_TO_KEY_SPACE)
        chips_on_board = list(machine.get_existing_xys_on_board(
            machine.get_chip_at(placement.x, placement.y)))

        # write how many chips to read
        spec.write_value(len(chips_on_board))

        # write the broad cast keys for timeouts
        router_timeout_key = (
            FecDataView.get_system_multicast_router_timeout_keys())
        # pylint: disable=unsubscriptable-object
        reinjection_base_key = router_timeout_key[(placement.x, placement.y)]
        spec.write_value(reinjection_base_key)

        mc_data_chips_to_keys = (
            FecDataView.get_data_in_multicast_key_to_chip_map())
        # write each chip x and y and base key
        for chip_xy in chips_on_board:
            board_chip_x, board_chip_y = machine.get_local_xy(
                machine.get_chip_at(*chip_xy))
            spec.write_value(board_chip_x)
            spec.write_value(board_chip_y)
            spec.write_value(mc_data_chips_to_keys[chip_xy])
            log.debug("for chip {}:{} base key is {}",
                      chip_xy[0], chip_xy[1], mc_data_chips_to_keys[chip_xy])

        # End-of-Spec:
        spec.end_specification()

    def _reserve_memory_regions(self, spec):
        """
        Writes the DSG regions memory sizes. Static so that it can be used
        by the application vertex.

        :param ~.DataSpecificationGenerator spec: spec file
        """
        spec.reserve_memory_region(
            region=_DATA_REGIONS.CONFIG,
            size=CONFIG_SIZE,
            label="config")
        spec.reserve_memory_region(
            region=_DATA_REGIONS.CHIP_TO_KEY_SPACE,
            size=SIZE_DATA_IN_CHIP_TO_KEY_SPACE,
            label="mc_key_map")
        spec.reserve_memory_region(
            region=_DATA_REGIONS.PROVENANCE_REGION,
            size=_PROVENANCE_DATA_SIZE, label="Provenance", empty=True)

    @overrides(AbstractHasAssociatedBinary.get_binary_file_name)
    def get_binary_file_name(self):
        return "data_speed_up_packet_gatherer.aplx"

    @staticmethod
    def locate_correct_write_data_function_for_chip_location(
            uses_advanced_monitors, x, y, transceiver,
            extra_monitor_cores_to_ethernet_connection_map):
        """
        Supports other components figuring out which gatherer and function
        to call for writing data onto SpiNNaker.

        :param bool uses_advanced_monitors:
            Whether the system is using advanced monitors
        :param int x: the chip x coordinate to write data to
        :param int y: the chip y coordinate to write data to
        :param ~spinnman.transceiver.Transceiver transceiver:
            the SpiNNMan instance
        :param extra_monitor_cores_to_ethernet_connection_map:
            mapping between cores and connections
        :type extra_monitor_cores_to_ethernet_connection_map:
            dict(tuple(int,int), DataSpeedUpPacketGatherMachineVertex)
        :return: a write function of either a LPG or the spinnMan
        :rtype: callable
        """
        if not uses_advanced_monitors:
            return transceiver.write_memory

        chip = FecDataView.get_chip_at(x, y)
        ethernet_connected_chip = FecDataView.get_chip_at(
            chip.nearest_ethernet_x, chip.nearest_ethernet_y)
        gatherer = extra_monitor_cores_to_ethernet_connection_map[
            ethernet_connected_chip.x, ethernet_connected_chip.y]
        return gatherer.send_data_into_spinnaker

    def _generate_data_in_report(
            self, time_diff, data_size, x, y,
            address_written_to, missing_seq_nums):
        """
        Writes the data in report for this stage.

        :param ~datetime.timedelta time_diff:
            the time taken to write the memory
        :param int data_size: the size of data that was written in bytes
        :param int x:
            the location in machine where the data was written to X axis
        :param int y:
            the location in machine where the data was written to Y axis
        :param int address_written_to: where in SDRAM it was written to
        :param list(set(int)) missing_seq_nums:
            the set of missing sequence numbers per data transmission attempt
        """
        if not os.path.isfile(self._in_report_path):
            with open(self._in_report_path, "w", encoding="utf-8") as writer:
                writer.write(
                    "x\t\t y\t\t SDRAM address\t\t size in bytes\t\t\t"
                    " time took \t\t\t Mb/s \t\t\t missing sequence numbers\n")
                writer.write(
                    "------------------------------------------------"
                    "------------------------------------------------"
                    "-------------------------------------------------\n")

        time_took_ms = float(time_diff.microseconds +
                             time_diff.total_seconds() * 1000000)
        megabits = (data_size * 8.0) / (1024 * BYTES_PER_KB)
        if time_took_ms == 0:
            mbs = "unknown, below threshold"
        else:
            mbs = megabits / (float(time_took_ms) / 100000.0)

        with open(self._in_report_path, "a", encoding="utf-8") as writer:
            writer.write(
                f"{x}\t\t {y}\t\t {address_written_to}\t\t {data_size}\t\t"
                f"\t\t {time_took_ms}\t\t\t {mbs}\t\t {missing_seq_nums}\n")

    def send_data_into_spinnaker(
            self, x, y, base_address, data, n_bytes=None, offset=0,
            cpu=0, is_filename=False):  # pylint: disable=unused-argument
        """
        Sends a block of data into SpiNNaker to a given chip.

        :param int x: chip x for data
        :param int y: chip y for data
        :param int base_address: the address in SDRAM to start writing memory
        :param data: the data to write (or filename to load data from,
            if ``is_filename`` is True; that's the only time this is a str)
        :type data: bytes or bytearray or memoryview or str
        :param int n_bytes: how many bytes to read, or `None` if not set
        :param int offset: where in the data to start from
        :param int cpu:
        :param bool is_filename: whether data is actually a file.
        """
        # if file, read in and then process as normal
        if is_filename:
            if offset != 0:
                raise ValueError(
                    "when using a file, you can only have a offset of 0")

            with open(data, "rb") as reader:
                # n_bytes=None already means 'read everything'
                data = reader.read(n_bytes)
            # Number of bytes to write is now length of buffer we have
            n_bytes = len(data)
        elif n_bytes is None:
            n_bytes = len(data)

        # start time recording
        start = datetime.datetime.now()
        # send data
        self._send_data_via_extra_monitors(
            x, y, base_address, data[offset:n_bytes + offset])
        # end time recording
        end = datetime.datetime.now()

        if VERIFY_SENT_DATA:
            original_data = bytes(data[offset:n_bytes + offset])
            transceiver = FecDataView.get_transceiver()
            verified_data = bytes(transceiver.read_memory(
                x, y, base_address, n_bytes))
            self.__verify_sent_data(
                original_data, verified_data, x, y, base_address, n_bytes)

        # write report
        if get_config_bool("Reports", "write_data_speed_up_reports"):
            self._generate_data_in_report(
                x=x, y=y, time_diff=end - start,
                data_size=n_bytes, address_written_to=base_address,
                missing_seq_nums=self._missing_seq_nums_data_in)

    @staticmethod
    def __verify_sent_data(
            original_data, verified_data, x, y, base_address, n_bytes):
        if original_data != verified_data:
            log.error("VARIANCE: chip:{},{} address:{} len:{}",
                      x, y, base_address, n_bytes)
            log.error("original:{}", original_data.hex())
            log.error("verified:{}", verified_data.hex())
            for i, (a, b) in enumerate(zip(original_data, verified_data)):
                if a != b:
                    raise ValueError(f"Mismatch found as position {i}")

    @staticmethod
    def __make_sdp_message(placement, port, payload):
        """
        :param ~.Placement placement:
        :param SDP_PORTS port:
        :param bytearray payload:
        :rtype: ~.SDPMessage
        """
        return SDPMessage(
            sdp_header=SDPHeader(
                destination_chip_x=placement.x,
                destination_chip_y=placement.y,
                destination_cpu=placement.p,
                destination_port=port.value,
                flags=SDPFlag.REPLY_NOT_EXPECTED),
            data=payload)

    def __open_connection(self):
        """
        Open an SCP connection and make our tag target it.

        :return: The opened connection, ready for use.
        :rtype: ~.SCAMPConnection
        """
        connection = open_scp_connection(self._x, self._y, self._ip_address)
        retarget_tag(connection, self._x, self._y, self._remote_tag)
        return connection

    def _send_data_via_extra_monitors(
            self, destination_chip_x, destination_chip_y, start_address,
            data_to_write):
        """
        Sends data using the extra monitor cores.

        :param int destination_chip_x: chip x
        :param int destination_chip_y: chip y
        :param int start_address: start address in SDRAM to write data to
        :param bytearray data_to_write: the data to write
        :param int start_address: the base SDRAM address
        """
        # Set up the connection
        with self.__open_connection() as self._connection:
            # how many packets after first one we need to send
            self._max_seq_num = ceildiv(
                len(data_to_write), BYTES_IN_FULL_PACKET_WITH_KEY)

            # determine board chip IDs, as the LPG does not know
            # machine scope IDs
            machine = FecDataView.get_machine()
            chip = FecDataView.get_chip_at(
                destination_chip_x, destination_chip_y)
            dest_x, dest_y = machine.get_local_xy(chip)
            self._coord_word = (dest_x << DEST_X_SHIFT) | dest_y

            # for safety, check the transaction id from the machine before
            # updating
            self.update_transaction_id_from_machine()
            self._transaction_id = (
                self._transaction_id + 1) & TRANSACTION_ID_CAP
            time_out_count = 0

            # verify completed
            received_confirmation = False
            while not received_confirmation:

                # send initial attempt at sending all the data
                self._send_all_data_based_packets(data_to_write, start_address)

                # Don't create a missing buffer until at least one packet has
                # come back.
                missing = None

                while not received_confirmation:
                    try:
                        # try to receive a confirmation of some sort from
                        # spinnaker
                        data = self._connection.receive(
                            timeout=self._TIMEOUT_PER_RECEIVE_IN_SECONDS)
                        time_out_count = 0

                        # Read command and transaction id
                        (cmd, transaction_id) = _TWO_WORDS.unpack_from(data, 0)

                        # If wrong transaction id, ignore packet
                        if self._transaction_id != transaction_id:
                            continue

                        # Decide what to do with the packet
                        if cmd == DATA_IN_COMMANDS.RECEIVE_FINISHED.value:
                            received_confirmation = True
                            break

                        if cmd != DATA_IN_COMMANDS.RECEIVE_MISSING_SEQ_DATA\
                                .value:
                            raise ValueError(f"Unknown command {cmd} received")

                        # The currently received packet has missing sequence
                        # numbers. Accumulate and dispatch transactionId when
                        # we've got them all.
                        if missing is None:
                            missing = set()
                            self._missing_seq_nums_data_in.append(missing)
                        seen_last, seen_all = self._read_in_missing_seq_nums(
                            data,
                            BYTES_FOR_RECEPTION_COMMAND_AND_ADDRESS_HEADER,
                            missing)

                        # Check that you've seen something that implies ready
                        # to retransmit.
                        if seen_all or seen_last:
                            self._outgoing_retransmit_missing_seq_nums(
                                data_to_write, missing)
                            missing.clear()

                    except SpinnmanTimeoutException as e:
                        # if the timeout has not occurred x times, keep trying
                        time_out_count += 1
                        if time_out_count > TIMEOUT_RETRY_LIMIT:
                            emergency_recover_state_from_failure(
                                self, self._placement)
                            raise SpinnFrontEndException(
                                "Failed to hear from the machine during "
                                f"{time_out_count} attempts. "
                                "Please try removing firewalls.") from e

                        # If we never received a packet, we will never have
                        # created the buffer, so send everything again
                        if missing is None:
                            break

                        self._outgoing_retransmit_missing_seq_nums(
                                data_to_write, missing)
                        missing.clear()

    def _read_in_missing_seq_nums(self, data, position, seq_nums):
        """
        Handles a missing sequence number packet from SpiNNaker.

        :param data: the data to translate into missing sequence numbers
        :type data: bytearray or bytes
        :param int position: the position in the data to write.
        :param set(int) seq_nums: a set of sequence numbers to add to
        :return: seen_last flag and seen_all flag
        :rtype: tuple(bool, bool)
        """
        # find how many elements are in this packet
        n_elements = (len(data) - position) // BYTES_PER_WORD

        # store missing
        new_seq_nums = n_word_struct(n_elements).unpack_from(
            data, position)

        # add missing seqs accordingly
        seen_last = False
        seen_all = False
        if new_seq_nums[-1] == self._MISSING_SEQ_NUMS_END_FLAG:
            new_seq_nums = new_seq_nums[:-1]
            seen_last = True
        if new_seq_nums[-1] == self.FLAG_FOR_MISSING_ALL_SEQUENCES:
            for missing_seq in range(0, self._max_seq_num):
                seq_nums.add(missing_seq)
            seen_all = True
        else:
            seq_nums.update(new_seq_nums)

        return seen_last, seen_all

    def _outgoing_retransmit_missing_seq_nums(
            self, data_to_write, missing):
        """
        Transmits back into SpiNNaker the missing data based off missing
        sequence numbers.

        :param bytearray data_to_write: the data to write.
        :param set(int) missing: a set of missing sequence numbers
        """

        missing_seqs_as_list = list(missing)
        missing_seqs_as_list.sort()

        # send seq data
        for missing_seq_num in missing_seqs_as_list:
            message, _length = self._calculate_data_in_data_from_seq_number(
                data_to_write, missing_seq_num,
                DATA_IN_COMMANDS.SEND_SEQ_DATA.value, None)
            self.__throttled_send(message)

        # request an update on what is missing
        self._send_tell_flag()

    @staticmethod
    def _calculate_position_from_seq_number(seq_num):
        """
        Calculates where in the raw data to start reading from, given a
        sequence number.

        :param int seq_num: the sequence number to determine position from
        :return: the position in the byte data
        :rtype: int
        """
        return BYTES_IN_FULL_PACKET_WITH_KEY * seq_num

    def _calculate_data_in_data_from_seq_number(
            self, data_to_write, seq_num, command_id, position):
        """
        Determine the data needed to be sent to the SpiNNaker machine
        given a sequence number.

        :param bytearray data_to_write:
            the data to write to the SpiNNaker machine
        :param int seq_num: the sequence number to get the data for
        :param int command_id:
        :param position: the position in the data to write to SpiNNaker
        :type position: int or None
        :return: SDP message and how much data has been written
        :rtype: tuple(~.SDPMessage, int)
        """

        # check for last packet
        packet_data_length = BYTES_IN_FULL_PACKET_WITH_KEY

        # determine position in data if not given
        if position is None:
            position = self._calculate_position_from_seq_number(seq_num)

        # if less than a full packet worth of data, adjust length
        if position + packet_data_length > len(data_to_write):
            packet_data_length = len(data_to_write) - position

        if packet_data_length < 0:
            raise ValueError("weird packet data length.")

        # determine the true packet length (with header)
        packet_length = (
            packet_data_length + BYTES_FOR_COMMAND_AND_ADDRESS_HEADER)

        # create struct
        packet_data = bytearray(packet_length)
        _THREE_WORDS.pack_into(
            packet_data, 0, command_id, self._transaction_id, seq_num)
        struct.pack_into(
            f"<{packet_data_length}B", packet_data,
            BYTES_FOR_COMMAND_AND_ADDRESS_HEADER,
            *data_to_write[position:position+packet_data_length])

        # debug
        # self._print_out_packet_data(packet_data, position)

        # build sdp packet
        message = self.__make_sdp_message(
            self._placement, SDP_PORTS.EXTRA_MONITOR_CORE_DATA_IN_SPEED_UP,
            packet_data)

        # return message for sending, and the length in data sent
        return message, packet_data_length

    def _send_location(self, start_address):
        """
        Send location as separate message.

        :param int start_address: SDRAM location
        """
        self._connection.send_sdp_message(self.__make_sdp_message(
            self._placement, SDP_PORTS.EXTRA_MONITOR_CORE_DATA_IN_SPEED_UP,
            _FIVE_WORDS.pack(
                DATA_IN_COMMANDS.SEND_DATA_TO_LOCATION.value,
                self._transaction_id, start_address, self._coord_word,
                self._max_seq_num - 1)))
        log.debug(
            "start address for transaction {} is {}",
            self._transaction_id, start_address)

    def _send_tell_flag(self):
        """
        Send tell flag as separate message.
        """
        self._connection.send_sdp_message(self.__make_sdp_message(
            self._placement, SDP_PORTS.EXTRA_MONITOR_CORE_DATA_IN_SPEED_UP,
            _TWO_WORDS.pack(
                DATA_IN_COMMANDS.SEND_TELL.value, self._transaction_id)))

    def _send_all_data_based_packets(self, data_to_write, start_address):
        """
        Send all the data as one block.

        :param bytearray data_to_write: the data to send
        :param int start_address:
        """
        # Send the location
        self._send_location(start_address)

        # where in the data we are currently up to
        position_in_data = 0

        # send rest of data
        for seq_num in range(0, self._max_seq_num):
            # put in command flag and seq num
            message, length_to_send = \
                self._calculate_data_in_data_from_seq_number(
                    data_to_write, seq_num,
                    DATA_IN_COMMANDS.SEND_SEQ_DATA.value, position_in_data)
            position_in_data += length_to_send

            # send the message
            self.__throttled_send(message)
            log.debug("sent seq {} of {} bytes", seq_num, length_to_send)

        # check for end flag
        self._send_tell_flag()
        log.debug("sent end flag")

    def set_cores_for_data_streaming(self):
        """
        Helper method for setting the router timeouts to a state usable
        for data streaming.
        """
        lead_monitor = FecDataView.get_monitor_by_xy(0, 0)
        # Store the last reinjection status for resetting
        # NOTE: This assumes the status is the same on all cores
        self._last_status = lead_monitor.get_reinjection_status()

        # Set to not inject dropped packets
        lead_monitor.set_reinjection_packets(
            point_to_point=False, multicast=False, nearest_neighbour=False,
            fixed_route=False)

        # Clear any outstanding packets from reinjection
        self.clear_reinjection_queue()

        # set time outs
        self.set_router_wait2_timeout(self._SHORT_TIMEOUT)
        self.set_router_wait1_timeout(self._LONG_TIMEOUT)

    @staticmethod
    def load_application_routing_tables():
        """
        Set all chips to have application table loaded in the router.
        """
        FecDataView.get_monitor_by_xy(0, 0).load_application_mc_routes()

    @staticmethod
    def load_system_routing_tables():
        """
        Set all chips to have the system table loaded in the router.
        """
        FecDataView.get_monitor_by_xy(0, 0).load_system_mc_routes()

    def set_router_wait1_timeout(self, timeout):
        """
        Set the wait1 field for a set of routers.

        :param tuple(int,int) timeout:
        :param ~pacman.model.placements.Placements placements:
        """
        mantissa, exponent = timeout
        core_subsets = convert_vertices_to_core_subset([self])
        process = SetRouterTimeoutProcess(
            FecDataView.get_scamp_connection_selector())
        try:
            process.set_wait1_timeout(mantissa, exponent, core_subsets)
        except:  # noqa: E722
            emergency_recover_state_from_failure(
                self, FecDataView.get_placement_of_vertex(self))
            raise

    def set_router_wait2_timeout(self, timeout):
        """
        Set the wait2 field for a set of routers.

        :param tuple(int,int) timeout:
        """
        mantissa, exponent = timeout
        core_subsets = convert_vertices_to_core_subset([self])
        process = SetRouterTimeoutProcess(
            FecDataView.get_scamp_connection_selector())
        try:
            process.set_wait2_timeout(mantissa, exponent, core_subsets)
        except:  # noqa: E722
            emergency_recover_state_from_failure(
                self, FecDataView.get_placement_of_vertex(self))
            raise

    def clear_reinjection_queue(self):
        """
        Clears the queues for reinjection.

        :param ~pacman.model.placements.Placements placements:
            the placements object
        """
        core_subsets = convert_vertices_to_core_subset([self])
        process = ClearQueueProcess(
            FecDataView.get_scamp_connection_selector())
        try:
            process.reset_counters(core_subsets)
        except:  # noqa: E722
            emergency_recover_state_from_failure(
                self, FecDataView.get_placement_of_vertex(self))
            raise

    def unset_cores_for_data_streaming(self):
        """
        Helper method for restoring the router timeouts to normal after
        being in a state usable for data streaming.
        """
        # Set the routers to temporary values
        self.set_router_wait1_timeout(self._TEMP_TIMEOUT)
        self.set_router_wait2_timeout(self._ZERO_TIMEOUT)

        if self._last_status is None:
            log.warning(
                "Cores have not been set for data extraction, so can't be"
                " unset")
        try:
            self.set_router_wait1_timeout(
                self._last_status.router_wait1_timeout_parameters)
            self.set_router_wait2_timeout(
                self._last_status.router_wait2_timeout_parameters)

            lead_monitor = FecDataView.get_monitor_by_xy(0, 0)
            lead_monitor.set_reinjection_packets(
                point_to_point=self._last_status.is_reinjecting_point_to_point,
                multicast=self._last_status.is_reinjecting_multicast,
                nearest_neighbour=(
                    self._last_status.is_reinjecting_nearest_neighbour),
                fixed_route=self._last_status.is_reinjecting_fixed_route)
        except Exception:  # pylint: disable=broad-except
            log.exception("Error resetting timeouts")
            log.error("Checking if the cores are OK...")
            core_subsets = convert_vertices_to_core_subset(
                FecDataView.iterate_monitors())
            try:
                transceiver = FecDataView.get_transceiver()
                error_cores = transceiver.get_cores_not_in_state(
                    core_subsets, {CPUState.RUNNING})
                if error_cores:
                    log.error("Cores in an unexpected state: {}", error_cores)
            except Exception:  # pylint: disable=broad-except
                log.exception("Couldn't get core state")

    def get_data(
            self, extra_monitor, placement, memory_address,
            length_in_bytes):
        """
        Gets data from a given core and memory address.

        :param ExtraMonitorSupportMachineVertex extra_monitor:
            the extra monitor used for this data
        :param ~pacman.model.placements.Placement placement:
            placement object for where to get data from
        :param int memory_address: the address in SDRAM to start reading from
        :param int length_in_bytes: the length of data to read in bytes
        :return: byte array of the data
        :rtype: bytearray
        """
        # create report elements
        if get_config_bool("Reports", "write_data_speed_up_reports"):
            routers_been_in_use = self._determine_which_routers_were_used(
                placement)
            self._write_routers_used_into_report(
                routers_been_in_use, placement)

        start = float(time.time())
        # if asked for no data, just return a empty byte array
        if length_in_bytes == 0:
            data = bytearray(0)
            end = float(time.time())
            with ProvenanceWriter() as db:
                # TODO Why log the time to not read???
                db.insert_gatherer(
                    placement.x, placement.y, memory_address, length_in_bytes,
                    self._run, "No Extraction time", end - start)
            return data

        transceiver = FecDataView.get_transceiver()

        # Update the IP Tag to work through a NAT firewall
        with self.__open_connection() as connection:
            # update transaction id for extra monitor
            extra_monitor.update_transaction_id()
            transaction_id = extra_monitor.transaction_id

            # send
            connection.send_sdp_message(self.__make_sdp_message(
                placement, SDP_PORTS.EXTRA_MONITOR_CORE_DATA_SPEED_UP,
                _FOUR_WORDS.pack(
                    DATA_OUT_COMMANDS.START_SENDING.value, transaction_id,
                    memory_address, length_in_bytes)))

            # receive
            self._output = bytearray(length_in_bytes)
            self._view = memoryview(self._output)
            self._max_seq_num = self.calculate_max_seq_num()
            lost_seq_nums = self._receive_data(
                transceiver, placement, connection, transaction_id)

            # Stop anything else getting through (and reduce traffic)
            connection.send_sdp_message(self.__make_sdp_message(
                placement, SDP_PORTS.EXTRA_MONITOR_CORE_DATA_SPEED_UP,
                _TWO_WORDS.pack(
                    DATA_OUT_COMMANDS.CLEAR.value, transaction_id)))

        end = float(time.time())
        with ProvenanceWriter() as db:
            db.insert_gatherer(
                placement.x, placement.y, memory_address, length_in_bytes,
                self._run, "Extraction time", end - start)
            for lost_seq_num in lost_seq_nums:
                if lost_seq_num > _MINOR_LOSS_THRESHOLD:
                    db.insert_report(
                        f"During the extraction of data of {length_in_bytes} "
                        f"bytes from memory address {memory_address} on "
                        f"chip ({placement.x}, {placement.y}), "
                        f"{lost_seq_num} sequences were lost.")
                if lost_seq_num > 0:
                    db.insert_gatherer(
                        placement.x, placement.y, memory_address,
                        length_in_bytes, self._run, "Lost_seq_nums",
                        lost_seq_num)

        return self._output

    def _receive_data(
            self, transceiver, placement, connection, transaction_id):
        """
        :param ~.Transceiver transceiver:
        :param ~.Placement placement:
        :param ~.UDPConnection connection:
        :param int transaction_id:
        :rtype: list(int)
        """
        seq_nums = set()
        lost_seq_nums = list()
        timeoutcount = 0
        finished = False
        while not finished:
            try:
                data = connection.receive(
                    timeout=self._TIMEOUT_PER_RECEIVE_IN_SECONDS)
                response_transaction_id, = _ONE_WORD.unpack_from(data, 4)
                if transaction_id == response_transaction_id:
                    timeoutcount = 0
                    seq_nums, finished = self._process_data(
                        data, seq_nums, finished, placement, transceiver,
                        lost_seq_nums, transaction_id)
                else:
                    log.info(
                        "ignoring packet as transaction id should be {}"
                        " but is {}", transaction_id, response_transaction_id)
            except SpinnmanTimeoutException as e:
                if timeoutcount > TIMEOUT_RETRY_LIMIT:
                    raise SpinnFrontEndException(
                        "Failed to hear from the machine during "
                        f"{timeoutcount} attempts. "
                        "Please try removing firewalls") from e

                timeoutcount += 1
                # self.__reset_connection()
                if not finished:
                    finished = self._determine_and_retransmit_missing_seq_nums(
                        seq_nums, transceiver, placement, lost_seq_nums,
                        transaction_id)
        return lost_seq_nums

    @staticmethod
    def _determine_which_routers_were_used(placement):
        """
        Traverse the fixed route paths from a given location to its
        destination. Used for determining which routers were used.

        :param ~.Placement placement: the source to start from
        :return: list of chip locations
        :rtype: list(tuple(int,int))
        """
        routers = [(placement.x, placement.y)]
        fixed_routes = FecDataView.get_fixed_routes()
        # pylint: disable=unsubscriptable-object
        entry = fixed_routes[placement.x, placement.y]
        chip_x = placement.x
        chip_y = placement.y
        while len(entry.processor_ids) == 0:
            # can assume one link, as its a minimum spanning tree going to
            # the root
            machine_link = FecDataView.get_chip_at(
                chip_x, chip_y).router.get_link(next(iter(entry.link_ids)))
            chip_x = machine_link.destination_x
            chip_y = machine_link.destination_y
            routers.append((chip_x, chip_y))
            entry = fixed_routes[chip_x, chip_y]
        return routers

    def _write_routers_used_into_report(self, routers_been_in_use, placement):
        """
        Write the used routers into a report.

        :param str report_path: the path to the report file
        :param list(tuple(int,int)) routers_been_in_use:
            the routers been in use
        :param ~.Placement placement: the first placement used
        """
        with open(self._out_report_path, "a", encoding="utf-8") as writer:
            writer.write(
                f"[{placement.x}:{placement.y}:{placement.p}] "
                f"= {routers_been_in_use}\n")

    def _calculate_missing_seq_nums(self, seq_nums):
        """
        Determine which sequence numbers we've missed.

        :param set(int) seq_nums: the set already acquired
        :return: list of missing sequence numbers
        :rtype: list(int)
        """
        return [sn for sn in range(self._max_seq_num) if sn not in seq_nums]

    def _determine_and_retransmit_missing_seq_nums(
            self, seq_nums, transceiver, placement, lost_seq_nums,
            transaction_id):
        """
        Determine if there are any missing sequence numbers, and if so
        retransmits the missing sequence numbers back to the core for
        retransmission.

        :param set(int) seq_nums: the sequence numbers already received
        :param ~.Transceiver transceiver: spinnman instance
        :param ~.Placement placement: placement instance
        :param list(int) lost_seq_nums:
        :param int transaction_id: transaction_id
        :return: whether all packets are transmitted
        :rtype: bool
        """
        # locate missing sequence numbers from pile
        missing_seq_nums = self._calculate_missing_seq_nums(seq_nums)

        lost_seq_nums.append(len(missing_seq_nums))
        # self._print_missing(missing_seq_nums)
        if not missing_seq_nums:
            return True

        # figure n packets given the 2 formats
        n_packets = 1
        length_via_format2 = len(missing_seq_nums) - (
            WORDS_PER_FULL_PACKET - WORDS_FOR_COMMAND_N_MISSING_TRANSACTION)
        if length_via_format2 > 0:
            n_packets += ceildiv(
                length_via_format2,
                WORDS_PER_FULL_PACKET - WORDS_FOR_COMMAND_TRANSACTION)
        # self._print_missing_n_packets(n_packets)

        # transmit missing sequence as a new SDP packet
        first = True
        seq_num_offset = 0
        for _ in range(n_packets):
            length_left_in_packet = WORDS_PER_FULL_PACKET
            offset = 0

            # if first, add n packets to list
            if first:

                # get left over space / data size
                size_of_data_left_to_transmit = min(
                    length_left_in_packet -
                    WORDS_FOR_COMMAND_N_MISSING_TRANSACTION,
                    len(missing_seq_nums) - seq_num_offset)

                # build data holder accordingly
                data = bytearray(
                    (size_of_data_left_to_transmit +
                     WORDS_FOR_COMMAND_N_MISSING_TRANSACTION) * BYTES_PER_WORD)

                # pack flag and n packets
                _ONE_WORD.pack_into(
                    data, offset, DATA_OUT_COMMANDS.START_MISSING_SEQ.value)
                _TWO_WORDS.pack_into(
                    data, BYTES_PER_WORD, transaction_id, n_packets)

                # update state
                offset += (
                    WORDS_FOR_COMMAND_N_MISSING_TRANSACTION * BYTES_PER_WORD)
                length_left_in_packet -= (
                    WORDS_FOR_COMMAND_N_MISSING_TRANSACTION)
                first = False

            else:  # just add data
                # get left over space / data size
                size_of_data_left_to_transmit = min(
                    WORDS_PER_FULL_PACKET_WITH_SEQUENCE_NUM,
                    len(missing_seq_nums) - seq_num_offset)

                # build data holder accordingly
                data = bytearray(
                    (size_of_data_left_to_transmit +
                     WORDS_FOR_COMMAND_TRANSACTION) * BYTES_PER_WORD)

                # pack flag
                _TWO_WORDS.pack_into(
                    data, offset,
                    DATA_OUT_COMMANDS.MISSING_SEQ.value, transaction_id)
                offset += BYTES_PER_WORD * WORDS_FOR_COMMAND_TRANSACTION
                length_left_in_packet -= WORDS_FOR_COMMAND_TRANSACTION

            # fill data field
            n_word_struct(size_of_data_left_to_transmit).pack_into(
                data, offset, *missing_seq_nums[
                    seq_num_offset:
                    seq_num_offset + size_of_data_left_to_transmit])
            seq_num_offset += length_left_in_packet

            # build SDP message and send it to the core
            transceiver.send_sdp_message(self.__make_sdp_message(
                placement, SDP_PORTS.EXTRA_MONITOR_CORE_DATA_SPEED_UP, data))

            # sleep for ensuring core doesn't lose packets
            time.sleep(self._TIMEOUT_FOR_SENDING_IN_SECONDS)
            # self._print_packet_num_being_sent(packet_count, n_packets)
        return False

    def _process_data(
            self, data, seq_nums, finished, placement, transceiver,
            lost_seq_nums, transaction_id):
        """
        Take a packet and process it see if we're finished yet.

        :param bytearray data: the packet data
        :param set(int) seq_nums: the list of sequence numbers received so far
        :param bool finished: bool which states if finished or not
        :param ~.Placement placement:
            placement object for location on machine
        :param ~.Transceiver transceiver: spinnman instance
        :param int transaction_id: the transaction ID for this stream
        :param list(int) lost_seq_nums:
            the list of n sequence numbers lost per iteration
        :return: set of data items, if its the first packet, the list of
            sequence numbers, the sequence number received and if its finished
        :rtype: tuple(set(int), bool)
        """
        # pylint: disable=too-many-arguments
        # self._print_out_packet_data(data)
        length_of_data = len(data)
        first_packet_element, = _ONE_WORD.unpack_from(data, 0)

        # get flags
        seq_num = first_packet_element & self._SEQUENCE_NUMBER_MASK
        is_end_of_stream = (
            first_packet_element & self._LAST_MESSAGE_FLAG_BIT_MASK) != 0

        # check seq num not insane
        if seq_num > self._max_seq_num:
            raise ValueError(
                f"got an insane sequence number. got {seq_num} when "
                f"the max is {self._max_seq_num} "
                f"with a length of {length_of_data}")

        # figure offset for where data is to be put
        offset = self._calculate_offset(seq_num)

        # write data

        # read offset from data is at byte 8. as first 4 is seq num,
        # second 4 is transaction id
        true_data_length = (
                offset + length_of_data - BYTES_FOR_SEQ_AND_TRANSACTION_ID)
        if (not is_end_of_stream or
                length_of_data != BYTES_FOR_SEQ_AND_TRANSACTION_ID):
            self._write_into_view(
                offset, true_data_length, data,
                BYTES_FOR_SEQ_AND_TRANSACTION_ID, length_of_data)

        # add seq num to list
        seq_nums.add(seq_num)

        # if received a last flag on its own, its during retransmission.
        #  check and try again if required
        if is_end_of_stream:
            if not self._check(seq_nums):
                finished = self._determine_and_retransmit_missing_seq_nums(
                    placement=placement, transceiver=transceiver,
                    seq_nums=seq_nums, lost_seq_nums=lost_seq_nums,
                    transaction_id=transaction_id)
            else:
                finished = True
        return seq_nums, finished

    @staticmethod
    def _calculate_offset(seq_num):
        """
        :param int seq_num:
        :rtype: int
        """
        return (seq_num * WORDS_PER_FULL_PACKET_WITH_SEQUENCE_NUM *
                BYTES_PER_WORD)

    def _write_into_view(
            self, view_start_position, view_end_position,
            data, data_start_position, data_end_position):
        """
        Puts data into the view.

        :param int view_start_position: where in view to start
        :param int view_end_position: where in view to end
        :param bytearray data: the data holder to write from
        :param int data_start_position: where in data holder to start from
        :param int data_end_position: where in data holder to end
        :param int seq_num: the sequence number to figure
        :raises Exception: If the position to write to is crazy
        """
        # pylint: disable=too-many-arguments
        if view_end_position > len(self._output):
            raise ValueError(
                f"End position {view_end_position} > "
                f"output length {len(self._output)}")
        self._view[view_start_position: view_end_position] = \
            data[data_start_position:data_end_position]

    def _check(self, seq_nums):
        """
        Verify if the sequence numbers are correct.

        :param list(int) seq_nums: the received sequence numbers
        :return: Whether all the sequence numbers have been received
        :rtype: bool
        """
        # hand back
        seq_nums = sorted(seq_nums)
        max_needed = self.calculate_max_seq_num()
        if len(seq_nums) > max_needed + 1:
            raise ValueError(f"too many seg_nums: {len(seq_nums)}")
        return len(seq_nums) == max_needed + 1

    def calculate_max_seq_num(self):
        """
        Deduce the max sequence number expected to be received.

        :return: the biggest sequence number expected
        :rtype: int
        """
        return ceildiv(
            len(self._output),
            WORDS_PER_FULL_PACKET_WITH_SEQUENCE_NUM * BYTES_PER_WORD)

    @staticmethod
    def _print_missing(seq_nums):
        """
        Debug printer for the missing sequence numbers from the pile.

        :param list(int) seq_nums: the sequence numbers received so far
        """
        for seq_num in sorted(seq_nums):
            log.info("from list I'm missing sequence num {}", seq_num)

    @staticmethod
    def _print_missing_n_packets(n_packets):
        """
        Debug printer for the number of missing packets from the pile.

        :param n_packets: the number of packets
        """
        log.info("missing packets = {}", n_packets)

    @staticmethod
    def _print_out_packet_data(data, position):
        """
        Debug prints out the data from the packet.

        :param bytearray data: the packet data
        """
        reread_data = struct.unpack(f"<{len(data)}B", data)
        output = ""
        position2 = position
        log.debug("size of data is {}", (len(data) / 4) - 3)
        for index, reread_data_element in enumerate(reread_data):
            if index >= 12:
                output += f"{position2}:{reread_data_element},"
                position2 += 1
        log.debug("converted data back into readable form is {}", output)

    @staticmethod
    def _print_length_of_received_seq_nums(seq_nums, max_needed):
        """
        Debug helper method for figuring out if everything been received.

        :param list(int) seq_nums: sequence numbers received
        :param int max_needed: biggest expected to have
        """
        if len(seq_nums) != max_needed:
            log.info("should have received {} sequence numbers, but received "
                     "{} sequence numbers", max_needed, len(seq_nums))

    @staticmethod
    def _print_packet_num_being_sent(packet_count, n_packets):
        """
        Debug helper for printing missing sequence number packet transmission.

        :param int packet_count: which packet is being fired
        :param int n_packets: how many packets to fire.
        """
        log.debug("send SDP packet with missing sequence numbers: {} of {}",
                  packet_count + 1, n_packets)

    @overrides(AbstractProvidesProvenanceDataFromMachine
               .get_provenance_data_from_machine)
    def get_provenance_data_from_machine(self, placement):
        # Get the App Data for the core
        transceiver = FecDataView.get_transceiver()
        region_table_address = transceiver.get_cpu_information_from_core(
            placement.x, placement.y, placement.p).user[0]

        # Get the provenance region base address
        prov_region_entry_address = get_region_base_address_offset(
            region_table_address, _DATA_REGIONS.PROVENANCE_REGION)
        provenance_address = transceiver.read_word(
            placement.x, placement.y, prov_region_entry_address)
        data = transceiver.read_memory(
            placement.x, placement.y, provenance_address,
            _PROVENANCE_DATA_SIZE)
        n_sdp_sent, n_sdp_recvd, n_in_streams, n_out_streams = (
            _FOUR_WORDS.unpack_from(data))
        with ProvenanceWriter() as db:
            db.insert_core(
                placement.x, placement.y, placement.p,
                "Sent_SDP_Packets", n_sdp_sent)
            db.insert_core(
                placement.x, placement.y, placement.p,
                "Received_SDP_Packets", n_sdp_recvd)
            db.insert_core(
                placement.x, placement.y, placement.p,
                "Speed_Up_Input_Streams", n_in_streams)
            db.insert_core(
                placement.x, placement.y, placement.p,
                "Speed_Up_Output_Streams", n_out_streams)



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
from spinn_utilities.progress_bar import ProgressBar
from spinn_machine.json_machine import to_json
from pacman.utilities import file_format_schemas
from spinn_front_end_common.data import FecDataView

MACHINE_FILENAME = "machine.json"


def write_json_machine(json_folder=None, progress_bar=True, validate=True):
    """
    Runs the code to write the machine in Java readable JSON.

    .. warning::
         The file in this folder will be overwritten!

    :param str json_folder: the folder to which the JSON are being written
    :param bool progress_bar: Flag if Progress Bar should be shown
    :param bool validate: Flag to disable the validation.
    :return: the name of the generated file
    :rtype: str
    """
    if progress_bar:
        # Steps are tojson, validate and writefile
        progress = ProgressBar(3, "Converting to JSON machine")
    else:
        progress = None
    if json_folder is None:
        json_folder = FecDataView.get_json_dir_path()
    file_path = os.path.join(json_folder, MACHINE_FILENAME)
    if not os.path.exists(file_path):
        json_obj = to_json()

        if progress:
            progress.update()

        if validate:
            # validate the schema
            file_format_schemas.validate(json_obj, MACHINE_FILENAME)

        # update and complete progress bar
        if progress:
            progress.end()

        # dump to json file
        with open(file_path, "w", encoding="utf-8") as f:
            json.dump(json_obj, f)

    if progress:
        progress.end()

    return file_path



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import struct
from spinn_utilities.config_holder import get_config_bool
from spinn_utilities.log import FormatAdapter
from spinn_utilities.progress_bar import ProgressBar
from spinn_machine import CoreSubsets, Router
from spinnman.model import ExecutableTargets
from spinnman.model.enums import CPUState
from spinn_front_end_common.data import FecDataView
from spinn_front_end_common.utilities.exceptions import SpinnFrontEndException
from spinn_front_end_common.utilities.system_control_logic import (
    run_system_application)
from spinn_front_end_common.utilities.utility_objs import ExecutableType
from spinn_front_end_common.utilities.helpful_functions import (
    get_defaultable_source_id)
from spinn_front_end_common.utilities.constants import COMPRESSOR_SDRAM_TAG
_FOUR_WORDS = struct.Struct("<IIII")
_THREE_WORDS = struct.Struct("<III")

logger = FormatAdapter(logging.getLogger(__name__))


def pair_compression():
    """
    Load routing tables and compress then using the Pair Algorithm.

    See ``pacman/operations/router_compressors/pair_compressor.py`` which is
    the exact same algorithm implemented in Python.

    :raises SpinnFrontEndException: If compression fails
     """
    # pylint: disable=too-many-arguments
    binary_path = FecDataView.get_executable_path(
        "simple_pair_compressor.aplx")
    compression = Compression(
        binary_path,
        "Running pair routing table compression on chip", result_register=1)
    compression.compress()


def ordered_covering_compression():
    """
    Load routing tables and compress then using the unordered Algorithm.

    To the best of our knowledge this is the same algorithm as
    :py:func:`mundy_on_chip_router_compression`, except this one is still
    buildable and can be maintained.

    :raises SpinnFrontEndException: If compression fails
    """
    # pylint: disable=too-many-arguments
    binary_path = FecDataView.get_executable_path(
        "simple_unordered_compressor.aplx")
    compression = Compression(
        binary_path,
        "Running unordered routing table compression on chip",
        result_register=1)
    compression.compress()


class Compression(object):
    """
    Compression algorithm implementation that uses a on-chip router
    compressor in order to parallelise.
    """

    __slots__ = [
        "_binary_path",
        "_compress_as_much_as_possible",
        "_compress_only_when_needed",
        "_compressor_app_id",
        "_progresses_text",
        "__result_register",
        "_routing_tables",
        "__failures"]

    def __init__(
            self, binary_path,  progress_text, result_register):
        """
        :param str binary_path: What binary to run
        :param ~spinn_machine.Machine machine: The machine model
        :param str progress_text: Text to use in progress bar
        :param int result_register:
            number of the user register to check for the result code
        """
        self._binary_path = binary_path
        self._compress_as_much_as_possible = get_config_bool(
            "Mapping", "router_table_compress_as_far_as_possible")
        # Only used by mundy compressor we can not rebuild
        self._compress_only_when_needed = None
        self._routing_tables = FecDataView.get_precompressed()
        self._progresses_text = progress_text
        self._compressor_app_id = None
        self.__failures = []
        self.__result_register = result_register

    def compress(self):
        """
        Apply the on-machine compression algorithm.

        :raises SpinnFrontEndException: If compression fails
        """
        # pylint: disable=too-many-arguments

        # build progress bar
        progress_bar = ProgressBar(
            len(self._routing_tables.routing_tables) * 2,
            self._progresses_text)

        if len(self._routing_tables.routing_tables) == 0:
            progress_bar.end()
            return

        self._compressor_app_id = FecDataView().get_new_id()

        # figure size of SDRAM needed for each chip for storing the routing
        # table
        for routing_table in progress_bar.over(self._routing_tables, False):
            self._load_routing_table(routing_table)

        # load the router compressor executable
        executable_targets = self._load_executables()

        run_system_application(
            executable_targets, self._compressor_app_id,
            get_config_bool("Reports", "write_compressor_iobuf"),
            self._check_for_success,
            [CPUState.FINISHED], False, "compressor_on_{}_{}_{}.txt",
            [self._binary_path], progress_bar)
        if self.__failures:
            raise SpinnFrontEndException(
                f"The router compressor failed on {self.__failures}")

    def _load_routing_table(self, table):
        """
        :param pacman.model.routing_tables.AbstractMulticastRoutingTable table:
            the pacman router table instance
        """
        transceiver = FecDataView.get_transceiver()
        data = self._build_data(table)

        # go to spinnman and ask for a memory region of that size per chip.
        base_address = transceiver.malloc_sdram(
            table.x, table.y, len(data), self._compressor_app_id,
            COMPRESSOR_SDRAM_TAG)

        # write SDRAM requirements per chip
        transceiver.write_memory(table.x, table.y, base_address, data)

    def _check_for_success(self, executable_targets):
        """
        Goes through the cores checking for cores that have failed to compress
        the routing tables to the level where they fit into the router.

        :param ~spinnman.model.ExecutableTargets executable_targets:
        """
        transceiver = FecDataView.get_transceiver()
        for core_subset in executable_targets.all_core_subsets:
            x = core_subset.x
            y = core_subset.y
            for p in core_subset.processor_ids:
                # Read the result from specified register
                if self.__result_register == 0:
                    result = transceiver.read_user_0(x, y, p)
                elif self.__result_register == 1:
                    result = transceiver.read_user_1(x, y, p)
                elif self.__result_register == 2:
                    result = transceiver.read_user_2(x, y, p)
                else:
                    raise ValueError(
                        "Incorrect register {self.__result_register}")
                # The result is 0 if success, otherwise failure
                if result != 0:
                    self.__failures.append((x, y))
        return len(self.__failures) == 0

    def _load_executables(self):
        """
        Plans the loading of the router compressor onto the chips.

        :return:
            the executable targets that represent all cores/chips which have
            active routing tables
        :rtype: ~spinnman.model.ExecutableTargets
        """
        # build core subsets
        core_subsets = CoreSubsets()
        for routing_table in self._routing_tables:
            # get the first none monitor core
            chip = FecDataView.get_chip_at(routing_table.x, routing_table.y)
            processor = chip.get_first_none_monitor_processor()

            # add to the core subsets
            core_subsets.add_processor(
                routing_table.x, routing_table.y, processor.processor_id)

        # build executable targets
        executable_targets = ExecutableTargets()
        executable_targets.add_subsets(self._binary_path, core_subsets,
                                       ExecutableType.SYSTEM)

        return executable_targets

    def _build_data(self, table):
        """
        Convert the router table into the data needed by the router
        compressor C code.

        :param table: the PACMAN router table instance
        :type table: ~pacman.model.routing_tables.AbstractMulticastRoutingTable
        :return: The byte array of data
        :rtype: bytearray
        """
        # write header data of the app ID to load the data, if to store
        # results in SDRAM and the router table entries

        data = b''
        if self._compress_only_when_needed is None:
            data += _THREE_WORDS.pack(
                FecDataView.get_app_id(),
                int(self._compress_as_much_as_possible),
                # Write the size of the table
                table.number_of_entries)
        else:
            # Mundy's compressor can not be changed so uses it own structure
            data += _FOUR_WORDS.pack(
                FecDataView.get_app_id(),
                int(self._compress_only_when_needed),
                int(self._compress_as_much_as_possible),
                # Write the size of the table
                table.number_of_entries)

        for entry in table.multicast_routing_entries:
            data += _FOUR_WORDS.pack(
                entry.routing_entry_key, entry.mask,
                Router.convert_routing_table_entry_to_spinnaker_route(entry),
                get_defaultable_source_id(entry))
        return bytearray(data)



# Copyright (c) 2019 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
import sys
from spinn_front_end_common.utilities.utility_objs import ExecutableType
from spinn_front_end_common.abstract_models import AbstractHasAssociatedBinary


class FindApplicationChipsUsed(object):
    """
    Builds a set of stats on how many chips were used for application cores.
    """

    def __call__(self, placements):
        """
        Finds how many application chips there were and the cost on each chip

        :param ~pacman.model.placements.Placements placements: placements
        :return: a tuple with 4 elements.

            1. how many chips were used
            2. the max application cores on any given chip
            3. the lowest number of application cores on any given chip
            4. the average number of application cores on any given chip

        :rtype: tuple(int,int,int,float)
        """
        chips_used = defaultdict(int)
        for placement in placements:
            # find binary type if applicable
            binary_start_type = None
            if isinstance(placement.vertex, AbstractHasAssociatedBinary):
                binary_start_type = placement.vertex.get_binary_start_type()
            if binary_start_type != ExecutableType.SYSTEM:
                chips_used[placement.x, placement.y] += 1

        low = sys.maxsize
        high = 0
        total = 0
        n_chips_used = len(chips_used)

        for key in chips_used:
            if chips_used[key] < low:
                low = chips_used[key]
            if chips_used[key] > high:
                high = chips_used[key]
            total += chips_used[key]
        average = total / n_chips_used
        return n_chips_used, high, low, average



# Copyright (c) 2016 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_utilities.progress_bar import ProgressBar
from spinn_front_end_common.data import FecDataView
from spinn_front_end_common.interface.provenance import (
    AbstractProvidesLocalProvenanceData)


def graph_provenance_gatherer():
    """
    Gets provenance information from the graph.
    """
    progress = ProgressBar(
        FecDataView.get_n_vertices() +
        FecDataView.get_n_partitions(),
        "Getting provenance data from application graph")
    for vertex in progress.over(FecDataView.iterate_vertices(), False):
        if isinstance(vertex, AbstractProvidesLocalProvenanceData):
            vertex.get_local_provenance_data()
            for m_vertex in vertex.machine_vertices:
                if isinstance(m_vertex, AbstractProvidesLocalProvenanceData):
                    m_vertex.get_local_provenance_data()
    for partition in progress.over(
            FecDataView.iterate_partitions()):
        for edge in partition.edges:
            if isinstance(edge, AbstractProvidesLocalProvenanceData):
                edge.get_local_provenance_data()



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import numpy
from spinn_utilities.progress_bar import ProgressBar
from spinn_machine import SDRAM
from data_specification import DataSpecificationExecutor
from data_specification.constants import MAX_MEM_REGIONS
from spinn_front_end_common.utilities.exceptions import SpinnFrontEndException
from spinn_front_end_common.utilities.helpful_functions import (
    get_region_base_address_offset)
from spinn_front_end_common.utilities.utility_calls import (
    get_data_spec_and_file_writer_filename)
from spinn_front_end_common.abstract_models import (
    AbstractRewritesDataSpecification)
from spinn_front_end_common.data import FecDataView
from spinn_front_end_common.utilities.helpful_functions import (
    generate_unique_folder_name)
from spinn_front_end_common.utilities.constants import BYTES_PER_WORD


def reload_dsg_regions():
    """
    Reloads DSG regions where needed.
    """
    # build file paths for reloaded stuff
    data_dir = get_reload_data_dir()

    progress = ProgressBar(
        FecDataView.get_n_placements(), "Reloading data")
    for placement in progress.over(FecDataView.iterate_placemements()):
        # Generate the data spec for the placement if needed
        regenerate_data_spec(placement, data_dir)

    # App data directory can be removed as should be empty
    os.rmdir(data_dir)


def get_reload_data_dir():
    """
    Get a path in which reloaded data files can be written.

    :rtype: str
    """
    run_dir_path = FecDataView.get_run_dir_path()
    data_dir = generate_unique_folder_name(
        run_dir_path, "reloaded_data_regions", "")
    if not os.path.exists(data_dir):
        os.makedirs(data_dir)
    return data_dir


def regenerate_data_spec(placement, data_dir):
    """
    Regenerate a data specification for a placement.

    :param ~.Placement placement: The placement to regenerate
    :param str data_dir: A place to use to write data to
    :return: Whether the data was regenerated or not
    :rtype: bool
    """
    vertex = placement.vertex

    # If the vertex doesn't regenerate, skip
    if not isinstance(vertex, AbstractRewritesDataSpecification):
        return False

    # If the vertex doesn't require regeneration, skip
    if not vertex.reload_required():
        return False

    txrx = FecDataView.get_transceiver()

    # build the writers for the reports and data
    spec_file, spec = get_data_spec_and_file_writer_filename(
        placement.x, placement.y, placement.p, data_dir)

    # Execute the regeneration
    vertex.regenerate_data_specification(spec, placement)

    # execute the spec
    with open(spec_file, "rb") as spec_reader:
        data_spec_executor = DataSpecificationExecutor(
            spec_reader, SDRAM.max_sdram_found)
        data_spec_executor.execute()
    try:
        os.remove(spec_file)
    except Exception:  # pylint: disable=broad-except
        # Ignore the deletion of files as non-critical
        pass

    # Read the region table for the placement
    regions_base_address = txrx.get_cpu_information_from_core(
        placement.x, placement.y, placement.p).user[0]
    start_region = get_region_base_address_offset(regions_base_address, 0)
    ptr_table = _get_ptr_table(
        txrx, placement, regions_base_address, start_region)

    # Get the data spec executor pointer table; note that this will not
    # have the right pointers since not all regions will be regenerated
    # but the checksum and size will be updated there
    ds_ptr_table = data_spec_executor.get_pointer_table(0)
    ds_ptr_table_bytes = ds_ptr_table.view("uint8")

    # Get the offset of the checksum and the checksum + size to write
    chk_off = DataSpecificationExecutor.TABLE_TYPE.fields['checksum'][1]
    chk_size = DataSpecificationExecutor.TABLE_TYPE.itemsize - chk_off

    # Write the regions to the machine
    for i, region in enumerate(data_spec_executor.dsef.mem_regions):
        if region is not None and not region.unfilled:
            # Verify that the region hasn't grown bigger than it was
            new_size = ds_ptr_table[i]["n_words"]
            old_size = ptr_table[i]["n_words"]
            if old_size != -1 and new_size > old_size:
                raise SpinnFrontEndException(
                    f"Region {i} of {vertex} has grown from {old_size} bytes "
                    f"to {new_size} bytes and will overwrite the next region!")
            txrx.write_memory(
                placement.x, placement.y, ptr_table[i]["pointer"],
                region.region_data[:region.max_write_pointer])
            # Update the checksum and size of the region
            reg = get_region_base_address_offset(regions_base_address, i)
            byte_offset = (reg - start_region) + chk_off
            addr = reg + chk_off
            txrx.write_memory(
                placement.x, placement.y, addr,
                ds_ptr_table_bytes[byte_offset:], chk_size)
    vertex.set_reload_required(False)
    return True


def _get_ptr_table(txrx, placement, regions_base_address, start_region):
    """
    Read the pointer table.

    :param ~spinnman.transceiver.Transceiver txrx:
        The transceiver to read with
    :param ~pacman.model.placements.Placement placement:
        Where to read the pointer table from
    :param int regions_base_address: The start of memory for the given core
    :param int start_region: The address of the first region address
    :rtype: ~numpy.ndarray(
        ~data_specification.DataSpecificationExecutor.TABLE_TYPE)
    """
    # Read the pointer table from the machine
    table_size = get_region_base_address_offset(
        regions_base_address, MAX_MEM_REGIONS) - start_region
    ptr_table = numpy.frombuffer(txrx.read_memory(
            placement.x, placement.y, start_region, table_size),
        dtype=DataSpecificationExecutor.TABLE_TYPE)

    # Fill in the size of regions which have no size but which are valid
    fixed_ptr_table = numpy.zeros(
        MAX_MEM_REGIONS, dtype=DataSpecificationExecutor.TABLE_TYPE)
    fixed_ptr_table[:] = ptr_table[:]
    last_no_size_region = None
    for i in range(0, MAX_MEM_REGIONS):

        # If this region is a valid region
        if ptr_table[i]["pointer"] != 0:

            # If there is a previous region with no size, use this region
            # pointer to find the size of the last one
            if last_no_size_region is not None:
                fixed_ptr_table[last_no_size_region]["n_words"] = (
                    ptr_table[i]["pointer"] -
                    ptr_table[last_no_size_region]["pointer"]) / BYTES_PER_WORD
                last_no_size_region = None

            # If this region has no size, make it the next to find the size of
            if ptr_table[i]["n_words"] == 0:
                last_no_size_region = i

    # We can get here and have the last region have no size; this is hard
    # to fix since we don't know where the data for the next region is!
    if last_no_size_region is not None:
        fixed_ptr_table[last_no_size_region]["n_words"] = -1
    return fixed_ptr_table



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .abstract_receive_buffers_to_host import AbstractReceiveBuffersToHost
from .abstract_sends_buffers_from_host import AbstractSendsBuffersFromHost
from .sends_buffers_from_host_pre_buffered_impl import (
    SendsBuffersFromHostPreBufferedImpl)

__all__ = ["AbstractReceiveBuffersToHost", "AbstractSendsBuffersFromHost",
           "SendsBuffersFromHostPreBufferedImpl"]



# Copyright (c) 2016 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from spinn_utilities.abstract_base import abstractproperty
from spinn_utilities.overrides import overrides
from spinn_front_end_common.utilities.helpful_functions import (
    get_region_base_address_offset)
from .abstract_provides_provenance_data_from_machine import (
    AbstractProvidesProvenanceDataFromMachine)
from spinn_front_end_common.data import FecDataView
from spinn_front_end_common.utilities.constants import BYTES_PER_WORD
from spinn_front_end_common.utilities.helpful_functions import n_word_struct
from spinn_front_end_common.interface.provenance.provenance_writer import (
    ProvenanceWriter)


def add_name(names, name):
    """
    :param iterable(str) names:
    :param str name:
    :rtype: list(str)
    """
    new_names = list(names)
    new_names.append(name)
    return new_names


class ProvidesProvenanceDataFromMachineImpl(
        AbstractProvidesProvenanceDataFromMachine, allow_derivation=True):
    """
    An implementation that gets provenance data from a region of integers on
    the machine.
    """

    __slots__ = ()

    class PROVENANCE_DATA_ENTRIES(Enum):
        """
        Entries for the provenance data generated by models using provides
        provenance vertex.
        """
        #: The counter of transmission overflows
        TRANSMISSION_EVENT_OVERFLOW = 0
        #: The counter of the number of times the callback queue was overloaded
        CALLBACK_QUEUE_OVERLOADED = 1
        #: The counter of the number of times the DMA queue was overloaded
        DMA_QUEUE_OVERLOADED = 2
        #: The counter of the number of times the user event queue overloaded
        USER_QUEUE_OVERLOADED = 3
        #: Whether the timer tick has overrun at all at any point
        TIMER_TIC_HAS_OVERRUN = 4
        #: The counter of the number of times the timer tick overran
        MAX_NUMBER_OF_TIMER_TIC_OVERRUN = 5

    N_SYSTEM_PROVENANCE_WORDS = len(PROVENANCE_DATA_ENTRIES)

    _TIMER_TICK_OVERRUN = "Times_the_timer_tic_over_ran"
    _MAX_TIMER_TICK_OVERRUN = "Max_number_of_times_timer_tic_over_ran"
    _TIMES_DMA_QUEUE_OVERLOADED = "Times_the_dma_queue_was_overloaded"
    _TIMES_USER_QUEUE_OVERLOADED = "Times_the_user_queue_was_overloaded"
    _TIMES_TRANSMISSION_SPIKES_OVERRAN = \
        "Times_the_transmission_of_spikes_overran"
    _TIMES_CALLBACK_QUEUE_OVERLOADED = \
        "Times_the_callback_queue_was_overloaded"

    @abstractproperty
    def _provenance_region_id(self):
        """
        The index of the provenance region.

        :rtype: int
        """

    @abstractproperty
    def _n_additional_data_items(self):
        """
        The number of extra machine words of provenance that the model reports.

        :rtype: int
        """

    def reserve_provenance_data_region(self, spec):
        """
        :param ~data_specification.DataSpecificationGenerator spec:
            The data specification being written.
        """
        spec.reserve_memory_region(
            self._provenance_region_id,
            self.get_provenance_data_size(self._n_additional_data_items),
            label="Provenance", empty=True)

    @classmethod
    def get_provenance_data_size(cls, n_additional_data_items):
        """
        :param int n_additional_data_items:
        :rtype: int
        """
        return (
            (cls.N_SYSTEM_PROVENANCE_WORDS + n_additional_data_items)
            * BYTES_PER_WORD)

    def _get_provenance_region_address(self, transceiver, placement):
        """
        :param ~spinnman.transceiver.Transceiver transceiver:
        :param ~pacman.model.placements.Placement placement:
        :rtype: int
        """
        # Get the App Data for the core
        region_table_address = transceiver.get_cpu_information_from_core(
            placement.x, placement.y, placement.p).user[0]

        # Get the provenance region base address
        prov_region_entry_address = get_region_base_address_offset(
            region_table_address, self._provenance_region_id)
        return transceiver.read_word(
            placement.x, placement.y, prov_region_entry_address)

    def _read_provenance_data(self, placement):
        """
        :param ~pacman.model.placements.Placement placement:
        :rtype: iterable(int)
        """
        transceiver = FecDataView.get_transceiver()
        provenance_address = self._get_provenance_region_address(
            transceiver, placement)
        data = transceiver.read_memory(
            placement.x, placement.y, provenance_address,
            self.get_provenance_data_size(self._n_additional_data_items))
        return n_word_struct(
            self.N_SYSTEM_PROVENANCE_WORDS +
            self._n_additional_data_items).unpack_from(data)

    @staticmethod
    def _get_provenance_placement_description(placement):
        """
        :param ~pacman.model.placements.Placement placement:
        :returns:
            A descriptive (human-readable) label and the (x, y, p) coordinates
            for provenance items from the given placement.
        :rtype: tuple(str, int, int, int)
        """
        label = placement.vertex.label
        x = placement.x
        y = placement.y
        p = placement.p
        desc_label = f"{label} on {x},{y},{p}"
        return desc_label, x, y, p

    def parse_system_provenance_items(self, label, x, y, p, provenance_data):
        """
        Given some words of provenance data, convert the portion of them that
        describes the system provenance into proper provenance items.

        Called by
        :py:meth:`~spinn_front_end_common.interface.provenance.ProvidesProvenanceDataFromMachineImpl.parse_extra_provenance_items.get_provenance_data_from_machine`

        :param str label:
            A descriptive label for the vertex (derived from label and placed
            position) to be used for provenance error reporting to the user.
        :param int x: x coordinate of the chip where this core
        :param int y: y coordinate of the core where this core
        :param int p: virtual id of the core
        :param list(int) provenance_data:
        """
        (tx_overflow, cb_overload, dma_overload, user_overload, tic_overruns,
         tic_overrun_max) = provenance_data[:self.N_SYSTEM_PROVENANCE_WORDS]

        # save provenance data items
        with ProvenanceWriter() as db:
            db.insert_core(
                x, y, p, self._TIMES_TRANSMISSION_SPIKES_OVERRAN, tx_overflow)
            if tx_overflow != 0:
                db.insert_report(
                    f"The transmission buffer for {label} was blocked on "
                    f"{tx_overflow} occasions. "
                    f" This is often a sign that the system is experiencing "
                    f"back pressure from the communication fabric. "
                    "Please either: "
                    "1. spread the load over more cores, "
                    "2. reduce your peak transmission load, or "
                    "3. adjust your mapping algorithm.")

            db.insert_core(
                x, y, p, self._TIMES_CALLBACK_QUEUE_OVERLOADED, cb_overload)
            if cb_overload != 0:
                db.insert_report(
                    f"The callback queue for {label} overloaded on "
                    f"{cb_overload} occasions.  "
                    f"This is often a sign that the system is running "
                    "too quickly for the number of neurons per core. "
                    "Please increase the machine time step or "
                    "time_scale_factor "
                    "or decrease the number of neurons per core.")

            db.insert_core(
                x, y, p, self._TIMES_DMA_QUEUE_OVERLOADED, dma_overload)
            if dma_overload != 0:
                db.insert_report(
                    f"The DMA queue for {label} overloaded on {dma_overload} "
                    "occasions.  "
                    "This is often a sign that the system is running "
                    "too quickly for the number of neurons per core.  "
                    "Please increase the machine time step or "
                    "time_scale_factor "
                    "or decrease the number of neurons per core.")

            db.insert_core(
                x, y, p, self._TIMES_USER_QUEUE_OVERLOADED, user_overload)
            if user_overload != 0:
                db.insert_report(
                    f"The USER queue for {label} overloaded on "
                    f"{user_overload} occasions.  "
                    f"This is often a sign that the system is running too "
                    f"quickly for the number of neurons per core.  Please "
                    f"increase the machine time step or time_scale_factor "
                    "or decrease the number of neurons per core.")

            db.insert_core(
                x, y, p, self._TIMER_TICK_OVERRUN, tic_overruns)
            if tic_overruns != 0:
                db.insert_report(
                    f"A Timer tick callback in {label} was still executing "
                    f"when the next timer tick callback was fired off "
                    f"{tic_overruns} times.  "
                    f"This is a sign of the system being overloaded and "
                    f"therefore the results are likely incorrect.  Please "
                    f"increase the machine time step or time_scale_factor "
                    f"or decrease the number of neurons per core")

            db.insert_core(
                x, y, p, self._MAX_TIMER_TICK_OVERRUN, tic_overrun_max)
            if tic_overrun_max > 0:
                db.insert_report(
                    f"The timer for {label} fell behind by up to "
                    f"{tic_overrun_max} ticks.  This is a sign of the system "
                    f"being overloaded and therefore the results are likely "
                    f"incorrect. Please increase the machine time step or "
                    f"time_scale_factor "
                    f"or decrease the number of neurons per core")

    def _get_extra_provenance_words(self, provenance_data):
        """
        Gets the words of provenance data not used for system provenance.

        :param list(int) provenance_data:
        :rtype: list(int)
        """
        return provenance_data[self.N_SYSTEM_PROVENANCE_WORDS:]

    def parse_extra_provenance_items(self, label, x, y, p, provenance_data):
        # pylint: disable=unused-argument
        """
        Convert the remaining provenance words (those not in the standard set)
        into provenance items.

        Called by
        :py:meth:`~spinn_front_end_common.interface.provenance.ProvidesProvenanceDataFromMachineImpl.parse_extra_provenance_items.get_provenance_data_from_machine`

        :param str label:
            A descriptive label for the vertex (derived from label and placed
            position) to be used for provenance error reporting to the user.
        :param int x: x coordinate of the chip where this core
        :param int y: y coordinate of the core where this core
        :param int p: virtual id of the core
        :param list(int) provenance_data:
            The list of words of raw provenance data.
        """
        if self._n_additional_data_items:
            raise NotImplementedError(
                f"{self} provides {self._n_additional_data_items} but doesn't "
                "parse them")

    @overrides(
        AbstractProvidesProvenanceDataFromMachine.
        get_provenance_data_from_machine,
        extend_doc=False)
    def get_provenance_data_from_machine(self, placement):
        """
        Retrieve the provenance data.

        :param ~pacman.model.placements.Placement placement:
            Which vertex are we retrieving from, and where was it
        """
        provenance_data = self._read_provenance_data(placement)
        label, x, y, p = self._get_provenance_placement_description(placement)
        self.parse_system_provenance_items(
            label, x, y, p, provenance_data)
        self.parse_extra_provenance_items(
            label, x, y, p, self._get_extra_provenance_words(provenance_data))



# Copyright (c) 2016 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import IntEnum
import struct
from spinn_utilities.overrides import overrides
from pacman.model.graphs.machine import MachineVertex
from pacman.model.resources import ConstantSDRAM
from spinn_front_end_common.data import FecDataView
from spinn_front_end_common.interface.provenance import (
    ProvidesProvenanceDataFromMachineImpl, ProvenanceWriter)
from spinn_front_end_common.interface.simulation.simulation_utilities import (
    get_simulation_header_array)
from spinn_front_end_common.abstract_models import (
    AbstractGeneratesDataSpecification, AbstractHasAssociatedBinary)
from spinn_front_end_common.utilities.utility_objs import ExecutableType
from spinn_front_end_common.utilities.constants import (
    SYSTEM_BYTES_REQUIREMENT, SIMULATION_N_BYTES, BYTES_PER_WORD)
from spinn_front_end_common.utilities.exceptions import ConfigurationException

_ONE_SHORT = struct.Struct("<H")
_TWO_BYTES = struct.Struct("<BB")


class LivePacketGatherMachineVertex(
        MachineVertex, ProvidesProvenanceDataFromMachineImpl,
        AbstractGeneratesDataSpecification, AbstractHasAssociatedBinary):
    """
    Used to gather multicast packets coming from cores and stream them
    out to a receiving application on host. Only ever deployed on chips
    with a working Ethernet connection.
    """
    class _REGIONS(IntEnum):
        SYSTEM = 0
        CONFIG = 1
        PROVENANCE = 2

    #: Used to identify tags involved with the live packet gatherer.
    TRAFFIC_IDENTIFIER = "LPG_EVENT_STREAM"

    _N_ADDITIONAL_PROVENANCE_ITEMS = 4
    _CONFIG_SIZE = 15 * BYTES_PER_WORD
    _PROVENANCE_REGION_SIZE = 2 * BYTES_PER_WORD
    KEY_ENTRY_SIZE = 3 * BYTES_PER_WORD

    def __init__(self, lpg_params, app_vertex=None, label=None):
        """
        :param LivePacketGatherParameters lpg_params: The parameters object
        :param LivePacketGather app_vertex: The application vertex
        :param str label: An optional label
        """
        # inheritance
        super().__init__(
            label or lpg_params.label, app_vertex=app_vertex)

        # app specific data items
        self._lpg_params = lpg_params
        self._incoming_sources = list()

    def add_incoming_source(self, m_vertex, partition_id):
        """
        Add a machine vertex source incoming into this gatherer.

        :param ~pacman.model.graphs.machine.MachineVertex m_vertex:
            The source machine vertex
        :param str partition_id: The incoming partition id
        """
        self._incoming_sources.append((m_vertex, partition_id))

    @property
    @overrides(ProvidesProvenanceDataFromMachineImpl._provenance_region_id)
    def _provenance_region_id(self):
        return self._REGIONS.PROVENANCE

    @property
    @overrides(ProvidesProvenanceDataFromMachineImpl._n_additional_data_items)
    def _n_additional_data_items(self):
        return self._N_ADDITIONAL_PROVENANCE_ITEMS

    def _get_key_translation_sdram(self):
        if not self._lpg_params.translate_keys:
            return 0
        return len(self._incoming_sources) * self.KEY_ENTRY_SIZE

    @property
    @overrides(MachineVertex.sdram_required)
    def sdram_required(self):
        return ConstantSDRAM(
            self.get_sdram_usage() + self._get_key_translation_sdram())

    @property
    @overrides(MachineVertex.iptags)
    def iptags(self):
        return [self._lpg_params.get_iptag_resource()]

    @overrides(
        ProvidesProvenanceDataFromMachineImpl.parse_extra_provenance_items)
    def parse_extra_provenance_items(self, label, x, y, p, provenance_data):
        (lost, lost_payload, events, messages) = provenance_data

        with ProvenanceWriter() as db:
            db.insert_core(x, y, p, "lost_packets_without_payload", lost)
            if lost > 0:
                db.insert_report(
                    f"The {label} has lost {lost} packets which do "
                    "not have payloads during its execution. "
                    "Try increasing the machine time step or increasing the "
                    "time scale factor. If you are running in real time, "
                    "try reducing the number of vertices which are feeding "
                    "this live packet gatherer")

            db.insert_core(x, y, p, "lost_packets_with_payload", lost_payload)
            if lost_payload > 0:
                db.insert_report(
                    f"The {label} has lost {lost_payload} packets "
                    "which have payloads during its execution. "
                    "Try increasing the machine time step or increasing the "
                    "time scale factor. If you are running in real time, "
                    "try reducing the number of vertices which are feeding "
                    "this live packet gatherer")

            db.insert_core(x, y, p, "gathered_events", events)

            db.insert_core(x, y, p, "messages_sent_to_host", messages)

    @overrides(AbstractHasAssociatedBinary.get_binary_file_name)
    def get_binary_file_name(self):
        return 'live_packet_gather.aplx'

    @overrides(AbstractHasAssociatedBinary.get_binary_start_type)
    def get_binary_start_type(self):
        return ExecutableType.USES_SIMULATION_INTERFACE

    @overrides(
        AbstractGeneratesDataSpecification.generate_data_specification)
    def generate_data_specification(
            self, spec, placement):  # @UnusedVariable
        # pylint: disable=arguments-differ
        spec.comment("\n*** Spec for LivePacketGather Instance ***\n\n")

        # Construct the data images needed for the Neuron:
        self._reserve_memory_regions(spec)
        self._write_setup_info(spec)
        self._write_configuration_region(
            spec, FecDataView.get_tags().get_ip_tags_for_vertex(self))

        # End-of-Spec:
        spec.end_specification()

    def _reserve_memory_regions(self, spec):
        """
        Reserve SDRAM space for memory areas.

        :param ~.DataSpecificationGenerator spec:
        """
        spec.comment("\nReserving memory space for data regions:\n\n")

        # Reserve memory:
        spec.reserve_memory_region(
            region=self._REGIONS.SYSTEM,
            size=SIMULATION_N_BYTES, label='system')
        spec.reserve_memory_region(
            region=self._REGIONS.CONFIG,
            size=self._CONFIG_SIZE + self._get_key_translation_sdram(),
            label='config')
        self.reserve_provenance_data_region(spec)

    def _write_configuration_region(self, spec, iptags):
        """
        Write the configuration region to the spec.

        :param ~.DataSpecificationGenerator spec:
        :param iterable(~.IPTag) iptags:
            The set of IP tags assigned to the object
        :raise ConfigurationException: if `iptags` is empty
        :raise DataSpecificationException:
            when something goes wrong with the DSG generation
        """
        spec.switch_write_focus(region=self._REGIONS.CONFIG)

        spec.write_value(int(self._lpg_params.use_prefix))
        spec.write_value(self._lpg_params.key_prefix or 0)
        spec.write_value(self._lpg_params.prefix_type.value
                         if self._lpg_params.prefix_type else 0)
        spec.write_value(self._lpg_params.message_type.value
                         if self._lpg_params.message_type else 0)
        spec.write_value(self._lpg_params.right_shift)
        spec.write_value(int(self._lpg_params.payload_as_time_stamps))
        spec.write_value(int(self._lpg_params.use_payload_prefix))
        spec.write_value(self._lpg_params.payload_prefix or 0)
        spec.write_value(data=self._lpg_params.payload_right_shift)

        # SDP tag
        for iptag in iptags:
            spec.write_value(iptag.tag)
            spec.write_value(_ONE_SHORT.unpack(_TWO_BYTES.pack(
                iptag.destination_y, iptag.destination_x))[0])
            break
        else:
            raise ConfigurationException("no iptag provided")

        # number of packets to send per time stamp
        spec.write_value(self._lpg_params.number_of_packets_sent_per_time_step)

        # Received key mask
        spec.write_value(self._lpg_params.received_key_mask)

        # Translated key right shift
        spec.write_value(self._lpg_params.translated_key_right_shift)

        # Key Translation
        if not self._lpg_params.translate_keys:
            spec.write_value(0)
        else:
            routing_info = FecDataView.get_routing_infos()
            spec.write_value(len(self._incoming_sources))
            for vertex, partition_id in self._incoming_sources:
                r_info = routing_info.get_routing_info_from_pre_vertex(
                    vertex, partition_id)
                spec.write_value(r_info.key)
                spec.write_value(r_info.mask)
                spec.write_value(vertex.vertex_slice.lo_atom)

    def _write_setup_info(self, spec):
        """
        Write basic info to the system region.

        :param ~.DataSpecificationGenerator spec:
        """
        # Write this to the system region (to be picked up by the simulation):
        spec.switch_write_focus(region=self._REGIONS.SYSTEM)
        spec.write_array(get_simulation_header_array(
            self.get_binary_file_name()))

    @classmethod
    def get_sdram_usage(cls):
        """
        Get the SDRAM used by this vertex.

        :rtype: int
        """
        return (
            SYSTEM_BYTES_REQUIREMENT + cls._CONFIG_SIZE +
            cls.get_provenance_data_size(cls._N_ADDITIONAL_PROVENANCE_ITEMS))

    @property
    def params(self):
        return self._lpg_params



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from pacman.model.graphs.application import ApplicationVertex
from .lpg_splitter import LPGSplitter


class LivePacketGather(ApplicationVertex):
    """
    A vertex that gathers and forwards multicast packets to the host.
    """
    __slots__ = ["__params"]

    def __init__(self, params, label=None):
        """
        :param LivePacketGatherParameters params: The parameters object
        :param str label: An optional label
        """
        super(LivePacketGather, self).__init__(label)
        self.__params = params
        self.splitter = LPGSplitter()

    @property
    def n_atoms(self):
        return 0

    @property
    def params(self):
        return self.__params



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from spinn_utilities.config_holder import (
    add_default_cfg, clear_cfg_files)
from spinnman.config_setup import add_spinnman_cfg
from pacman.config_setup import add_pacman_cfg
from data_specification.config_setup import add_data_specification_cfg
from spinn_front_end_common.data.fec_data_writer import FecDataWriter

BASE_CONFIG_FILE = "spinnaker.cfg"


def unittest_setup():
    """
    Does all the steps that may be required before a unit test.

    Resets the configurations so only the local default configurations are
    included.

    Unsets any previous simulators and temporary directories.

    .. note::
        This file should only be called from `spinn_front_end_common/tests`
    """
    clear_cfg_files(True)
    add_spinnaker_cfg()
    FecDataWriter.mock()


def add_spinnaker_cfg():
    """
    Add the local configuration and all dependent configuration files.
    """
    add_pacman_cfg()  # This add its dependencies too
    add_spinnman_cfg()  # double adds of dependencies ignored
    add_data_specification_cfg()  # double adds of dependencies ignored
    add_default_cfg(os.path.join(os.path.dirname(__file__), BASE_CONFIG_FILE))



# Copyright (c) 2016 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from spinn_utilities.progress_bar import ProgressBar
from spinn_utilities.log import FormatAdapter
from spinnman.exceptions import SpinnmanException
from spinn_front_end_common.data import FecDataView
from spinn_front_end_common.interface.provenance import ProvenanceWriter

logger = FormatAdapter(logging.getLogger(__name__))


def router_provenance_gatherer():
    gather = _RouterProvenanceGatherer()
    # pylint: disable=protected-access
    gather._add_router_provenance_data()


class _RouterProvenanceGatherer(object):
    """
    Gathers diagnostics from the routers.
    """

    __slots__ = []

    def _add_router_provenance_data(self):
        """
        Writes the provenance data of the router diagnostics
        """
        progress = ProgressBar(FecDataView.get_machine().n_chips*2,
                               "Getting Router Provenance")

        seen_chips = set()

        # get all extra monitor core data if it exists
        reinjection_data = None
        if FecDataView.has_monitors():
            monitor = FecDataView.get_monitor_by_xy(0, 0)
            reinjection_data = monitor.get_reinjection_status_for_vertices()

        for router_table in progress.over(
                FecDataView.get_uncompressed().routing_tables, False):
            seen_chips.add(self._add_router_table_diagnostic(
                router_table, reinjection_data))

        # Get what info we can for chips where there are problems or no table
        for chip in progress.over(sorted(
                FecDataView.get_machine().chips, key=lambda c: (c.x, c.y))):
            if (chip.x, chip.y) not in seen_chips:
                self._add_unseen_router_chip_diagnostic(
                    chip, reinjection_data)

    def _add_router_table_diagnostic(self, table, reinjection_data):
        """
        :param ~.MulticastRoutingTable table:
        :param dict(tuple(int,int),ReInjectionStatus) reinjection_data:
        """
        x = table.x
        y = table.y
        try:
            transceiver = FecDataView.get_transceiver()
            diagnostics = transceiver.get_router_diagnostics(x, y)
        except SpinnmanException:
            logger.warning(
                "Could not read routing diagnostics from {}, {}",
                x, y, exc_info=True)
            return
        status = self.__get_status(reinjection_data, x, y)
        self.__router_diagnostics(x, y, diagnostics, status, True, table)
        return x, y

    def _add_unseen_router_chip_diagnostic(self, chip, reinjection_data):
        """
        :param ~.Chip chip:
        :param dict(tuple(int,int),ReInjectionStatus) reinjection_data:
        """
        try:
            transceiver = FecDataView.get_transceiver()
            diagnostics = transceiver.get_router_diagnostics(chip.x, chip.y)
        except SpinnmanException:
            # There could be issues with unused chips - don't worry!
            return
        if (diagnostics.n_dropped_multicast_packets or
                diagnostics.n_local_multicast_packets or
                diagnostics.n_external_multicast_packets):
            status = self.__get_status(reinjection_data, chip.x, chip.y)
            self.__router_diagnostics(
                chip.x, chip.y, diagnostics, status, False, None)

    @staticmethod
    def __get_status(reinjection_data, x, y):
        """
        :param dict(tuple(int,int),ReInjectionStatus) reinjection_data:
        :param int x:
        :param int y:
        :rtype: ReInjectionStatus or None
        """
        return reinjection_data[x, y] if reinjection_data else None

    def __router_diagnostics(self, x, y, diagnostics, status, expected, table):
        """
        Describes the router diagnostics for one router.

        :param int x: x coordinate of the router in question
        :param int y: y coordinate of the router in question
        :param ~.RouterDiagnostics diagnostics: the router diagnostics object
        :param ReInjectionStatus status:
            the data gained from the extra monitor re-injection subsystem
        :param bool expected:
        :param ~.AbstractMulticastRoutingTable table:
            the router table generated by the PACMAN tools
        """
        # pylint: disable=too-many-arguments

        # simplify the if by making components of it outside.
        has_dropped = (diagnostics.n_dropped_multicast_packets > 0)
        missing_stuff = False
        has_reinjection = status is not None
        if has_reinjection:
            missing_stuff = ((
                status.n_dropped_packets + status.n_missed_dropped_packets +
                status.n_dropped_packet_overflows +
                status.n_reinjected_packets + status.n_processor_dumps +
                status.n_link_dumps) < diagnostics.n_dropped_multicast_packets)

        with ProvenanceWriter() as db:
            db.insert_router(
                x, y, "Local_Multicast_Packets",
                diagnostics.n_local_multicast_packets, expected)

            db.insert_router(
                x, y, "External_Multicast_Packets",
                diagnostics.n_external_multicast_packets, expected)

            db.insert_router(
                x, y, "Dropped_Multicast_Packets",
                diagnostics.n_dropped_multicast_packets, expected)
            if (has_dropped and not has_reinjection) or (
                    has_dropped and has_reinjection and missing_stuff):
                db.insert_report(
                    f"The router on {x}, {y} has dropped "
                    f"{diagnostics.n_dropped_multicast_packets} "
                    f"multicast route packets. "
                    f"Try increasing the machine_time_step and/or the time "
                    f"scale factor or reducing the number of atoms per core.")

            db.insert_router(
                x, y, "Dropped_Multicast_Packets_via_local_transmission",
                diagnostics.user_3, expected)
            if diagnostics.user_3 > 0:
                db.insert_report(
                    f"The router on {x}, {y} has dropped {diagnostics.user_3} "
                    "multicast packets that were transmitted by local cores. "
                    "This occurs where the router has no entry associated "
                    "with the multicast key. "
                    "Try investigating the keys allocated to the vertices "
                    "and the router table entries for this chip.")

            db.insert_router(
                x, y, "default_routed_external_multicast_packets",
                diagnostics.user_2, expected)
            if diagnostics.user_2 > 0 and not (
                    table and table.number_of_defaultable_entries):
                db.insert_report(
                    f"The router on {x}, {y} has default routed "
                    f"{diagnostics.user_2} multicast packets, but the router "
                    f"table did not expect any default routed packets. "
                    f"This occurs where the router has no entry associated "
                    f"with the multicast key. "
                    f"Try investigating the keys allocated to the vertices "
                    f"and the router table entries for this chip.")

            if table:
                db.insert_router(
                    x, y, "Entries", table.number_of_entries, expected)
                routes = set()
                for ent in table.multicast_routing_entries:
                    routes.add(ent.spinnaker_route)
                db.insert_router(x, y, "Unique_Routes", len(routes), expected)

            db.insert_router(
                x, y, "Local_P2P_Packets",
                diagnostics.n_local_peer_to_peer_packets, expected)

            db.insert_router(
                x, y, "External_P2P_Packets",
                diagnostics.n_external_peer_to_peer_packets, expected)

            db.insert_router(
                x, y, "Dropped_P2P_Packets",
                diagnostics.n_dropped_peer_to_peer_packets, expected)

            db.insert_router(
                x, y, "Local_NN_Packets",
                diagnostics.n_local_nearest_neighbour_packets, expected)

            db.insert_router(
                x, y, "External_NN_Packets",
                diagnostics.n_external_nearest_neighbour_packets, expected)

            db.insert_router(
                x, y, "Dropped_NN_Packets",
                diagnostics.n_dropped_nearest_neighbour_packets, expected)

            db.insert_router(
                x, y, "Local_FR_Packets",
                diagnostics.n_local_fixed_route_packets, expected)

            db.insert_router(
                x, y, "External_FR_Packets",
                diagnostics.n_external_fixed_route_packets, expected)

            db.insert_router(
                x, y, "Dropped_FR_Packets",
                diagnostics.n_dropped_fixed_route_packets, expected)
            if diagnostics.n_dropped_fixed_route_packets > 0:
                db.insert_report(
                    f"The router on chip {x}:{y} dropped "
                    f"{diagnostics.n_dropped_fixed_route_packets} fixed "
                    f"route packets. "
                    f"This is indicative of an error within the data "
                    f"extraction process as this is the only expected user of "
                    "fixed route packets.")

            db.insert_router(
                x, y, "Error status", diagnostics.error_status, expected)
            if diagnostics.error_status > 0:
                db.insert_report(
                    f"The router on {x}, {y} has a non-zero error status. "
                    f"This could indicate a hardware fault. "
                    f"The errors set are {diagnostics.errors_set}, and the "
                    f"error count is {diagnostics.error_count}")

            if status is None:
                return  # rest depends on status

            db.insert_router(
                x, y, "Received_For_Reinjection",
                status.n_dropped_packets, expected)

            db.insert_router(
                x, y, "Missed_For_Reinjection",
                status.n_missed_dropped_packets, expected)
            if status.n_missed_dropped_packets > 0:
                db.insert_report(
                    f"The extra monitor on {x}, {y} has missed "
                    f"{status.n_missed_dropped_packets} packets.")

            db.insert_router(
                x, y, "Reinjection_Overflows",
                status.n_dropped_packet_overflows, expected,)
            if status.n_dropped_packet_overflows > 0:
                db.insert_report(
                    f"The extra monitor on {x}, {y} has dropped "
                    f"{status.n_dropped_packet_overflows} packets.")

            db.insert_router(
                x, y, "Reinjected", status.n_reinjected_packets, expected)

            db.insert_router(
                x, y, "Dumped_from_a_Link", status.n_link_dumps, expected)
            if status.n_link_dumps > 0:
                db.insert_report(
                    f"The extra monitor on {x}, {y} has detected that "
                    f"{status.n_link_dumps} packets were dumped from "
                    f"outgoing links {status.links_dropped_from} of this "
                    f"chip's router. This often occurs "
                    f"when external devices are used in the script but not "
                    f"connected to the communication fabric correctly. "
                    f"These packets may have been reinjected multiple times "
                    f"and so this number may be an overestimate.")

            db.insert_router(
                x, y, "Dumped_from_a_processor", status.n_processor_dumps,
                expected)
            if status.n_processor_dumps > 0:
                db.insert_report(
                    f"The extra monitor on {x}, {y} has detected that "
                    f"{status.n_processor_dumps} packets were dumped from "
                    f"cores {status.processors_dropped_from} failing to take "
                    "the packet. This often occurs when "
                    "the executable has crashed or has not been given a "
                    "multicast packet callback. It can also result from the "
                    "core taking too long to process each packet. These "
                    "packets were reinjected and so this number is likely an "
                    "overestimate.")



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
from spinn_utilities.progress_bar import ProgressBar
from spinn_front_end_common.data import FecDataView
from spinn_front_end_common.utilities.report_functions import reports

_FOLDER_NAME = "routing_tables_from_machine"


def routing_table_from_machine_report(routing_tables):
    """
    Report the routing table that was actually on the machine.

        folder_name = os.path.join(FecDataView().run_dir_path, _FOLDER_NAME)
        os.mkdir(folder_name)

    :param routing_tables: Compressed routing tables
    :type routing_tables:
        ~pacman.model.routing_tables.MulticastRoutingTables
    :param ~spinnman.transceiver.Transceiver transceiver:
    :param int app_id:
    """
    tables = list(routing_tables.routing_tables)
    progress = ProgressBar(tables, "Reading Routing Tables from Machine")

    folder_name = os.path.join(FecDataView.get_run_dir_path(), _FOLDER_NAME)
    os.mkdir(folder_name)

    # generate a file for every multicast entry
    for routing_table in progress.over(tables):
        reports.generate_routing_table(routing_table, folder_name)



# Copyright (c) 2022 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import auto, Enum


class TimerWork(Enum):
    """
    Different Work types an Algorithm could be doing
    """

    OTHER = (auto(), "Other")
    GET_MACHINE = (auto(), "Turning on Machine")
    LOADING = (auto(), "Loading Stage")
    # LOADING
    BITFIELD = (auto(), "BitField work")
    # Only for on Machine Compression
    COMPRESSING = (auto(), "Compressing")
    CONTROL = (auto(), "Control")
    SYNAPSE = (auto(), "Expanding Synapse")
    RUNNING = (auto(), "Running")
    EXTRACTING = (auto(), "Extracting")
    # TODO is in right to treat this separately
    EXTRACT_DATA = (auto(), "Extracting")
    REPORT = (auto(), "Reporting")

    def __new__(cls, value, work_name):
        # pylint: disable=protected-access
        obj = object.__new__(cls)
        obj._value_ = value
        obj._work_name = work_name
        return obj

    @property
    def work_name(self):
        return self._work_name



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .read_status_process import ReadStatusProcess
from .reset_counters_process import ResetCountersProcess
from .set_packet_types_process import SetPacketTypesProcess
from .set_router_timeout_process import SetRouterTimeoutProcess
from .clear_queue_process import ClearQueueProcess
from .load_application_mc_routes_process import LoadApplicationMCRoutesProcess
from .load_system_mc_routes_process import LoadSystemMCRoutesProcess

__all__ = [
    "ClearQueueProcess",
    "LoadApplicationMCRoutesProcess",
    "LoadSystemMCRoutesProcess",
    "ReadStatusProcess",
    "ResetCountersProcess",
    "SetPacketTypesProcess",
    "SetRouterTimeoutProcess",
    ]



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from spinn_utilities.progress_bar import ProgressBar
from spinnman.messages.scp.enums import Signal
from spinnman.model import ExecutableTargets
from spinnman.model.enums import CPUState
from spinn_front_end_common.data import FecDataView
from spinn_front_end_common.utilities.helpful_functions import (
    flood_fill_binary_to_spinnaker)
from spinn_front_end_common.utilities.utility_objs import ExecutableType
from spinnman.exceptions import SpiNNManCoresNotInStateException
from spinn_front_end_common.utilities.emergency_recovery import (
    emergency_recover_states_from_failure)

# 10 seconds is lots of time to wait for the application to become ready!
_APP_READY_TIMEOUT = 10.0


def load_app_images():
    """
    Go through the executable targets and load each binary to everywhere
    and then send a start request to the cores that actually use it.
    """
    __load_images(lambda ty: ty is not ExecutableType.SYSTEM,
                  "Loading executables onto the machine")


def load_sys_images():
    """
    Go through the executable targets and load each binary to everywhere
    and then send a start request to the cores that actually use it.
    """
    __load_images(lambda ty: ty is ExecutableType.SYSTEM,
                  "Loading system executables onto the machine")
    try:
        _, cores = filter_targets(lambda ty: ty is ExecutableType.SYSTEM)
        FecDataView.get_transceiver().wait_for_cores_to_be_in_state(
            cores.all_core_subsets, FecDataView.get_app_id(),
            [CPUState.RUNNING], timeout=10)
    except SpiNNManCoresNotInStateException as e:
        emergency_recover_states_from_failure()
        raise e


def __load_images(filter_predicate, label):
    """
    :param callable(ExecutableType,bool) filter_predicate:
    :param str label
    """
    # Compute what work is to be done here
    binaries, cores = filter_targets(filter_predicate)

    try:
        # ISSUE: Loading order may be non-constant on older Python
        progress = ProgressBar(cores.total_processors + 1, label)
        for binary in binaries:
            progress.update(flood_fill_binary_to_spinnaker(binary))

        __start_simulation(cores, FecDataView.get_app_id())
        progress.update()
        progress.end()
    except Exception as e:
        try:
            FecDataView.get_transceiver().stop_application(
                FecDataView.get_app_id())
        except Exception:  # pylint: disable=broad-except
            # Ignore this, this was just an attempt at recovery
            pass
        raise e


def filter_targets(filter_predicate):
    """
    :param callable(ExecutableType,bool) filter_predicate:
    :rtype: tuple(list(str), ~spinnman.model.ExecutableTargets)
    """
    binaries = []
    cores = ExecutableTargets()
    targets = FecDataView.get_executable_targets()
    for exe_type in targets.executable_types_in_binary_set():
        if filter_predicate(exe_type):
            for aplx in targets.get_binaries_of_executable_type(exe_type):
                binaries.append(aplx)
                cores.add_subsets(
                    aplx, targets.get_cores_for_binary(aplx), exe_type)
    return binaries, cores


def __start_simulation(cores, app_id):
    """
    :param ~.ExecutableTargets cores:
        Possible subset of all ExecutableTargets to start
    :param int app_id:
    """
    txrx = FecDataView.get_transceiver()
    txrx.wait_for_cores_to_be_in_state(
        cores.all_core_subsets, app_id, [CPUState.READY],
        timeout=_APP_READY_TIMEOUT)
    txrx.send_signal(app_id, Signal.START)



# Copyright (c) 2021 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import logging
import math
import os
import time
from spinn_utilities.config_holder import (
    get_config_int, get_config_str)
from spinn_utilities.log import FormatAdapter
from spinn_utilities.overrides import overrides
from spinn_front_end_common.utilities.notification_protocol import (
    NotificationProtocol)
from spinnman.data.spinnman_data_writer import SpiNNManDataWriter
from spinnman.messages.scp.enums.signal import Signal
from spinnman.model import ExecutableTargets
from pacman.data.pacman_data_writer import PacmanDataWriter
from pacman.model.routing_tables import MulticastRoutingTables
from spinn_front_end_common.interface.buffer_management import BufferManager
from spinn_front_end_common.interface.ds import DsSqlliteDatabase
from spinn_front_end_common.interface.java_caller import JavaCaller
from spinn_front_end_common.utilities.constants import (
    MICRO_TO_MILLISECOND_CONVERSION, MICRO_TO_SECOND_CONVERSION)
from spinn_front_end_common.utilities.exceptions import ConfigurationException
from spinn_front_end_common.utility_models import (
    DataSpeedUpPacketGatherMachineVertex, ExtraMonitorSupportMachineVertex)
from spinn_front_end_common.abstract_models import (
    AbstractMachineAllocationController)
from .fec_data_view import FecDataView, _FecDataModel


logger = FormatAdapter(logging.getLogger(__name__))
__temp_dir = None

REPORTS_DIRNAME = "reports"
# pylint: disable=protected-access


class FecDataWriter(PacmanDataWriter, SpiNNManDataWriter, FecDataView):
    """
    See :py:class:`~spinn_utilities.data.utils_data_writer.UtilsDataWriter`.

    This class is designed to only be used directly by
    :py:class:`AbstractSpinnakerBase`
    and within the Non-PyNN repositories unit tests as all methods are
    available to subclasses.
    """
    __fec_data = _FecDataModel()
    __slots__ = []

    @overrides(PacmanDataWriter._mock)
    def _mock(self):
        PacmanDataWriter._mock(self)
        self._spinnman_mock()
        self.__fec_data._clear()
        # run numbers start at 1 and when not running this is the next one
        self.__fec_data._run_number = 1
        self.set_up_timings(1000, 1)

    @overrides(PacmanDataWriter._setup)
    def _setup(self):
        PacmanDataWriter._setup(self)
        self._spinnman_setup()
        self.__fec_data._clear()
        # run numbers start at 1 and when not running this is the next one
        self.__fec_data._run_number = 1
        self.__create_reports_directory()
        self.__create_timestamp_directory()
        self.__create_run_dir_path()

    @overrides(PacmanDataWriter.finish_run)
    def finish_run(self):
        PacmanDataWriter.finish_run(self)
        self.__fec_data._run_number += 1

    @overrides(PacmanDataWriter._hard_reset)
    def _hard_reset(self):
        if self.is_ran_last():
            self.__fec_data._reset_number += 1
        PacmanDataWriter._hard_reset(self)
        SpiNNManDataWriter._local_hard_reset(self)
        self.__fec_data._hard_reset()
        self.__create_run_dir_path()

    @overrides(PacmanDataWriter._soft_reset)
    def _soft_reset(self):
        if self.is_ran_last():
            self.__fec_data._reset_number += 1
        PacmanDataWriter._soft_reset(self)
        SpiNNManDataWriter._local_soft_reset(self)
        self.__fec_data._soft_reset()

    def __create_run_dir_path(self):
        self.set_run_dir_path(self._child_folder(
            self.__fec_data._timestamp_dir_path,
            f"run_{self.__fec_data._run_number}"))

    def __create_reports_directory(self):
        default_report_file_path = get_config_str(
            "Reports", "default_report_file_path")
        # determine common report folder
        if default_report_file_path == "DEFAULT":
            directory = os.getcwd()

            # global reports folder
            self.set_report_dir_path(
                self._child_folder(directory, REPORTS_DIRNAME))
        else:
            self.set_report_dir_path(
                self._child_folder(default_report_file_path, REPORTS_DIRNAME))

    def __create_timestamp_directory(self):
        while True:
            try:
                now = datetime.datetime.now()
                timestamp = (
                    f"{now.year:04}-{now.month:02}-{now.day:02}-{now.hour:02}"
                    f"-{now.minute:02}-{now.second:02}-{now.microsecond:06}")
                self.__fec_data._timestamp_dir_path = self._child_folder(
                    self.get_report_dir_path(), timestamp, must_create=True)
                return
            except OSError:
                time.sleep(0.5)

    def set_allocation_controller(self, allocation_controller):
        """
        Sets the allocation controller variable.

        :param AbstractMachineAllocationController allocation_controller:
        """
        if allocation_controller and not isinstance(
                allocation_controller, AbstractMachineAllocationController):
            raise TypeError(
                "allocation_controller must be a MachineAllocationController")
        self.__fec_data._allocation_controller = allocation_controller

    def set_buffer_manager(self, buffer_manager):
        """
        Sets the Buffer manager variable.

        :param BufferManager buffer_manager:
        """
        if not isinstance(buffer_manager, BufferManager):
            raise TypeError("buffer_manager must be a BufferManager")
        self.__fec_data._buffer_manager = buffer_manager

    def increment_current_run_timesteps(self, increment):
        """
        Increment the current_run_timesteps and sets first_machine_time_step.

        A `None` increment signals run_forever

        :param increment: The timesteps for this do_run loop
        :type increment: int or None
        """
        if increment is None:
            if self.__fec_data._current_run_timesteps != 0:
                raise NotImplementedError("Run forever after another run")
            self.__fec_data._current_run_timesteps = None
        else:
            if not isinstance(increment, int):
                raise TypeError("increment should be an int (or None")
            if increment < 0:
                raise ConfigurationException(
                    f"increment {increment} must not be negative")
            if self.__fec_data._current_run_timesteps is None:
                raise NotImplementedError("Run after run forever")
            self.__fec_data._first_machine_time_step = \
                self.__fec_data._current_run_timesteps
            self.__fec_data._current_run_timesteps += increment

    def set_max_run_time_steps(self, max_run_time_steps):
        """
        Sets the max_run_time_steps value

        :param int max_run_time_steps: new value
        """
        if not isinstance(max_run_time_steps, int):
            raise TypeError("max_run_time_steps should be an int")
        if max_run_time_steps <= 0:
            raise ConfigurationException(
                f"max_run_time_steps {max_run_time_steps} must be positive")
        self.__fec_data._max_run_time_steps = max_run_time_steps

    def set_up_timings(
            self, simulation_time_step_us, time_scale_factor,
            default_time_scale_factor=None):
        """
        Set up timings for the simulation.

        :param simulation_time_step_us:
            An explicitly specified time step for the simulation in .
            If `None`, the value is read from the configuration
        :type simulation_time_step_us: int or None
        :param time_scale_factor:
            An explicitly specified time scale factor for the simulation.
            If `None`, the value is read from the configuration
        :type time_scale_factor: float or None
        :param default_time_scale_factor:
            A back up time scale factor for the simulation.
            Only used if time_scale_factor parameter and configuration are
            both `None`.
            If `None`, the value is based on `simulation_time_step`
        :type default_time_scale_factor: float or None
        """
        try:
            self._set_simulation_time_step(simulation_time_step_us)
            self._set_time_scale_factor(
                time_scale_factor, default_time_scale_factor)
            self._set_hardware_timestep()
        except ConfigurationException:
            self.__fec_data._simulation_time_step_us = None
            self.__fec_data._simulation_time_step_ms = None
            self.__fec_data._simulation_time_step_per_ms = None
            self.__fec_data._simulation_time_step_per_s = None
            self.__fec_data._simulation_time_step_s = None
            self.__fec_data._time_scale_factor = None
            self.__fec_data._hardware_time_step_us = None
            self.__fec_data._hardware_time_step_ms = None
            raise

    def _set_simulation_time_step(self, simulation_time_step_us):
        """
        :param simulation_time_step_us:
            An explicitly specified time step for the simulation.  If `None`,
            the value is read from the configuration
        :type simulation_time_step: int or None
        """
        if simulation_time_step_us is None:
            simulation_time_step_us = get_config_int(
                "Machine", "simulation_time_step")

        if not isinstance(simulation_time_step_us, int):
            raise TypeError("simulation_time_step_us should be an int")

        if simulation_time_step_us <= 0:
            raise ConfigurationException(
                f'invalid simulation_time_step {simulation_time_step_us}'
                ': must greater than zero')

        self.__fec_data._simulation_time_step_us = simulation_time_step_us
        self.__fec_data._simulation_time_step_ms = (
                simulation_time_step_us / MICRO_TO_MILLISECOND_CONVERSION)
        self.__fec_data._simulation_time_step_per_ms = (
                MICRO_TO_MILLISECOND_CONVERSION / simulation_time_step_us)
        self.__fec_data._simulation_time_step_per_s = (
                MICRO_TO_SECOND_CONVERSION / simulation_time_step_us)
        self.__fec_data._simulation_time_step_s = (
                simulation_time_step_us / MICRO_TO_SECOND_CONVERSION)

    def _set_time_scale_factor(
            self, time_scale_factor, default_time_scale_factor):
        """
        Set up time_scale_factor.

        If time_scale_factor is provide that is used

        Then if configuration is not `None` that is used

        Then if default is provided that is used

        Lastly it is set based on the simulation_time_step

        :param time_scale_factor:
            An explicitly specified time scale factor for the simulation.
            If `None`, the value is read from the configuration
        :type time_scale_factor: float or None
        """
        if time_scale_factor is None:
            # Note while this reads from the cfg the cfg default is None
            time_scale_factor = get_config_int("Machine", "time_scale_factor")

        if time_scale_factor is None:
            if default_time_scale_factor is not None:
                time_scale_factor = default_time_scale_factor

        if time_scale_factor is None:
            time_scale_factor = max(
                1.0, math.ceil(self.get_simulation_time_step_per_ms()))
            if time_scale_factor > 1.0:
                logger.warning(
                    f"A timestep was entered that has forced spinnaker to "
                    f"automatically slow the simulation down from real time "
                    f"by a factor of {time_scale_factor}.")

        if not isinstance(time_scale_factor, (int, float)):
            raise TypeError("app_id should be an int (or float)")

        if time_scale_factor <= 0:
            raise ConfigurationException(
                f'invalid time_scale_factor {time_scale_factor}'
                ': must greater than zero')

        self.__fec_data._time_scale_factor = time_scale_factor

    def _set_hardware_timestep(self):
        raw = (self.get_simulation_time_step_us() *
               self.get_time_scale_factor())
        rounded = round(raw)
        if abs(rounded - raw) > 0.0001:
            raise ConfigurationException(
                "The multiplication of simulation time step in microseconds: "
                f"{self.get_simulation_time_step_us()} and times scale factor"
                f": {self.get_time_scale_factor()} produced a non integer "
                f"hardware time step of {raw}")

        logger.info(f"Setting hardware timestep as {rounded} microseconds "
                    f"based on simulation time step of "
                    f"{self.get_simulation_time_step_us()} and "
                    f"timescale factor of {self.get_time_scale_factor()}")
        self.__fec_data._hardware_time_step_us = rounded
        self.__fec_data._hardware_time_step_ms = (
                rounded / MICRO_TO_MILLISECOND_CONVERSION)

    def set_system_multicast_routing_data(self, data):
        """
        Sets the system_multicast_routing_data.

        These are: `data_in_multicast_routing_tables`,
        `data_in_multicast_key_to_chip_map`,
        `system_multicast_router_timeout_keys`

        :param data: new value
        :type data:
            tuple(~pacman.model.routing_tables.MulticastRoutingTables,
            dict, dict)
        """
        (data_in_multicast_routing_tables,
         data_in_multicast_key_to_chip_map,
         system_multicast_router_timeout_keys) = data
        if not isinstance(data_in_multicast_routing_tables,
                          MulticastRoutingTables):
            raise TypeError("First element must be a MulticastRoutingTables")
        if not isinstance(data_in_multicast_key_to_chip_map, dict):
            raise TypeError("Second element must be dict")
        if not isinstance(system_multicast_router_timeout_keys, dict):
            raise TypeError("Third element must be a dict")
        self.__fec_data._data_in_multicast_key_to_chip_map = \
            data_in_multicast_key_to_chip_map
        self.__fec_data._data_in_multicast_routing_tables = \
            data_in_multicast_routing_tables
        self.__fec_data._system_multicast_router_timeout_keys = \
            system_multicast_router_timeout_keys

    def set_n_required(self, n_boards_required, n_chips_required):
        """
        Sets (if not `None`) the number of boards/chips requested by the user.

        :param n_boards_required:
            `None` or the number of boards requested by the user
        :type n_boards_required: int or None
        :param n_chips_required:
            `None` or the number of chips requested by the user
        :type n_chips_required: int or None
        """
        if n_boards_required is None:
            if n_chips_required is None:
                return
            elif not isinstance(n_chips_required, int):
                raise TypeError("n_chips_required must be an int (or None)")
            if n_chips_required <= 0:
                raise ConfigurationException(
                    "n_chips_required must be positive and not "
                    f"{n_chips_required}")
        else:
            if n_chips_required is not None:
                raise ConfigurationException(
                    "Illegal call with both both param provided as "
                    f"{n_boards_required}, {n_chips_required}")
            if not isinstance(n_boards_required, int):
                raise TypeError("n_boards_required must be an int (or None)")
            if n_boards_required <= 0:
                raise ConfigurationException(
                    "n_boards_required must be positive and not "
                    f"{n_boards_required}")
        if self.__fec_data._n_boards_required is not None or \
                self.__fec_data._n_chips_required is not None:
            raise ConfigurationException(
                "Illegal second call to set_n_required")
        self.__fec_data._n_boards_required = n_boards_required
        self.__fec_data._n_chips_required = n_chips_required

    def set_n_chips_in_graph(self, n_chips_in_graph):
        """
        Sets the number of chips needed by the graph.

        :param int n_chips_in_graph:
        """
        if not isinstance(n_chips_in_graph, int):
            raise TypeError("n_chips_in_graph must be an int (or None)")
        if n_chips_in_graph <= 0:
            raise ConfigurationException(
                "n_chips_in_graph must be positive and not "
                f"{n_chips_in_graph}")
        self.__fec_data._n_chips_in_graph = n_chips_in_graph

    def set_ipaddress(self, ip_address):
        """
        :param str ip_address:
        """
        if not isinstance(ip_address, str):
            raise TypeError("ipaddress must be a str")
        self.__fec_data._ipaddress = ip_address

    def set_fixed_routes(self, fixed_routes):
        """
        :param fixed_routes:
        :type fixed_routes:
            dict(tuple(int,int), ~spinn_machine.FixedRouteEntry)
        """
        if not isinstance(fixed_routes, dict):
            raise TypeError("fixed_routes must be a dict")
        self.__fec_data._fixed_routes = fixed_routes

    def set_java_caller(self, java_caller):
        """
        :param JavaCaller java_caller:
        """
        if not isinstance(java_caller, JavaCaller):
            raise TypeError("java_caller must be a JavaCaller")
        self.__fec_data._java_caller = java_caller

    def reset_sync_signal(self):
        """
        Returns the sync signal to the default value.
        """
        self.__fec_data._next_sync_signal = Signal.SYNC0

    def set_executable_types(self, executable_types):
        """
        :param executable_types:
        :type executable_types: dict(
            ~spinn_front_end_common.utilities.utility_objs.ExecutableType,
            ~spinn_machine.CoreSubsets or None)
        """
        if not isinstance(executable_types, dict):
            raise TypeError("executable_types must be a Dict")
        self.__fec_data._executable_types = executable_types

    def set_live_packet_gatherer_parameters(self, params):
        """
        testing method will not work outside of mock
        """
        if not self._is_mocked():
            raise NotImplementedError("This call is only for testing")
        self.__fec_data._live_packet_recorder_params = params

    def set_database_file_path(self, database_file_path):
        """
        Sets the database_file_path variable. Possibly to `None`.

        :param database_file_path:
        :type database_file_path: str or None
        """
        if not isinstance(database_file_path, (str, type(None))):
            raise TypeError("database_file_path must be a str or None")
        self.__fec_data._database_file_path = database_file_path

    def set_executable_targets(self, executable_targets):
        """
        Sets the executable_targets

        :param ~spinnman.model.ExecutableTargets executable_targets:
        """
        if not isinstance(executable_targets, ExecutableTargets):
            raise TypeError("executable_targets must be a ExecutableTargets")
        self.__fec_data._executable_targets = executable_targets

    def set_dsg_targets(self, dsg_targets):
        """
        Sets the Data Spec targets database.

        :param ~spinnman.model.ExecutableTargets dsg_targets:
        """
        if not isinstance(dsg_targets, DsSqlliteDatabase):
            raise TypeError("dsg_targets must be a DsSqlliteDatabase")
        self.__fec_data._dsg_targets = dsg_targets

    def __gatherer_map_error(self):
        return TypeError(
            "gatherer_map must be a dict((int, int), "
            "DataSpeedUpPacketGatherMachineVertex)")

    def set_gatherer_map(self, gatherer_map):
        """
        Sets the map of (x,y) to Gatherer Vertices.

        :param gatherer_map:
        :type gatherer_map:
            dict((int, int), DataSpeedUpPacketGatherMachineVertex)
        """
        if not isinstance(gatherer_map, dict):
            raise self.__gatherer_map_error()
        try:
            for (x, y), vertex in gatherer_map.items():
                if not isinstance(x, int):
                    raise self.__gatherer_map_error()
                if not isinstance(y, int):
                    raise self.__gatherer_map_error()
                if not isinstance(
                        vertex, DataSpeedUpPacketGatherMachineVertex):
                    raise self.__gatherer_map_error()
                break  # assume if first is ok all are
        except Exception as ex:  # pylint: disable=broad-except
            raise self.__gatherer_map_error() from ex
        self.__fec_data._gatherer_map = gatherer_map

    def __monitor_map_error(self):
        return TypeError(
            "monitor_map must be a dict((int, int), "
            "ExtraMonitorSupportMachineVertex)")

    def set_monitor_map(self, monitor_map):
        """
        Sets the map of (x,y) to Monitor Vertices.

        :param monitor_map:
        :type monitor_map:
            dict((int, int), ExtraMonitorSupportMachineVertex)
        """
        if not isinstance(monitor_map, dict):
            raise self.__monitor_map_error()
        try:
            for (x, y), vertex in monitor_map.items():
                if not isinstance(x, int):
                    raise self.__monitor_map_error()
                if not isinstance(y, int):
                    raise self.__monitor_map_error()
                if not isinstance(vertex, ExtraMonitorSupportMachineVertex):
                    raise self.__monitor_map_error()
                break  # assume if first is ok all are
        except TypeError:
            raise
        except Exception as ex:  # pylint: disable=broad-except
            raise self.__monitor_map_error() from ex
        self.__fec_data._monitor_map = monitor_map

    def set_notification_protocol(self, notification_protocol):
        """
        Sets the notification_protocol.

        :param NotificationProtocol notification_protocol:
        """
        self.__fec_data._clear_notification_protocol()
        if not isinstance(notification_protocol, NotificationProtocol):
            raise TypeError(
                "notification_protocol must be a NotificationProtocol")
        self.__fec_data._notification_protocol = notification_protocol

    def clear_notification_protocol(self):
        """
        Closes an existing notification_protocol and sets the value to `None`.

        If no notification_protocol exist this method silently returns.

        If the close causes an Exception it is logged and ignored
        """
        self.__fec_data._clear_notification_protocol()

    @classmethod
    @overrides(FecDataView.add_vertex)
    def add_vertex(cls, vertex):
        # Avoid the safety check in FecDataView
        PacmanDataWriter.add_vertex(vertex)

    def next_run_step(self):
        """
        Starts or increases the run step count.

        Run steps start at 1

        :return: The next step number
        :rtype: int
        """
        if self.__fec_data._run_step is None:
            self.__fec_data._run_step = 1
        else:
            self.__fec_data._run_step += 1
        return self.__fec_data._run_step

    def clear_run_steps(self):
        """
        Clears the run step.

        get_run_step will go back to returning `None`

        next_run_step will restart at 1
        """
        self.__fec_data._run_step = None



# Copyright (c) 2019 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
from spinn_utilities.progress_bar import ProgressBar
from pacman.model.graphs.machine import SourceSegmentedSDRAMMachinePartition
from spinn_front_end_common.data import FecDataView
from spinn_front_end_common.utilities.exceptions import SpinnFrontEndException
from spinn_front_end_common.utilities.constants import SDRAM_EDGE_BASE_TAG


def sdram_outgoing_partition_allocator():
    if FecDataView.has_transceiver():
        transceiver = FecDataView.get_transceiver()
        virtual_usage = None
    else:
        # Ok if transceiver = None
        transceiver = None
        virtual_usage = defaultdict(int)

    progress_bar = ProgressBar(
        total_number_of_things_to_do=FecDataView.get_n_vertices(),
        string_describing_what_being_progressed=(
            "Allocating SDRAM for SDRAM outgoing egde partitions"))

    # Keep track of SDRAM tags used
    next_tag = defaultdict(lambda: SDRAM_EDGE_BASE_TAG)

    for vertex in FecDataView.iterate_vertices():
        sdram_partitions = vertex.splitter.get_internal_sdram_partitions()
        for sdram_partition in sdram_partitions:

            # get placement, ones where the src is multiple,
            # you need to ask for the first pre vertex
            if isinstance(
                    sdram_partition, SourceSegmentedSDRAMMachinePartition):
                placement = FecDataView.get_placement_of_vertex(
                    next(iter(sdram_partition.pre_vertices)))
            else:
                placement = FecDataView.get_placement_of_vertex(
                    sdram_partition.pre_vertex)

            # total sdram
            total_sdram = (sdram_partition.total_sdram_requirements())

            # if bust, throw exception
            if total_sdram == 0:
                raise SpinnFrontEndException(
                    "Cannot allocate SDRAM size of 0 for "
                    f"partition {sdram_partition}")

            # allocate
            if transceiver is not None:
                tag = next_tag[placement.x, placement.y]
                next_tag[placement.x, placement.y] = tag + 1
                sdram_base_address = transceiver.malloc_sdram(
                    placement.x, placement.y, total_sdram,
                    FecDataView.get_app_id(), tag)
            else:
                sdram_base_address = virtual_usage[
                    placement.x, placement.y]
                virtual_usage[placement.x, placement.y] += total_sdram

            # update
            sdram_partition.sdram_base_address = sdram_base_address

        progress_bar.update()
    progress_bar.end()



# Copyright (c) 2022 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import struct
import logging
from spinn_utilities.progress_bar import ProgressBar
from spinnman.messages.spinnaker_boot import SystemVariableDefinition
from spinn_utilities.config_holder import get_config_bool
from spinn_utilities.log import FormatAdapter
from spinn_front_end_common.data import FecDataView

# The fixed point position for drift readings
DRIFT_FP = 1 << 17
CLOCK_DRIFT_REPORT = "clock_drift.csv"

logger = FormatAdapter(logging.getLogger(__name__))


def drift_report():
    """
    A report on the clock drift as reported by each chip
    """
    ethernet_only = get_config_bool(
            "Reports", "drift_report_ethernet_only")
    machine = FecDataView.get_machine()
    eth_chips = machine.ethernet_connected_chips
    n_chips = machine.n_chips
    if ethernet_only:
        n_chips = len(eth_chips)

    # create file path
    directory_name = os.path.join(
        FecDataView.get_run_dir_path(), CLOCK_DRIFT_REPORT)

    # If the file is new, write a header
    if not os.path.exists(directory_name):
        with open(directory_name, "w", encoding="utf-8") as writer:
            for eth_chip in eth_chips:
                if ethernet_only:
                    writer.write(f'"{eth_chip.x} {eth_chip.y}",')
                else:
                    for chip in machine.get_chips_by_ethernet(
                            eth_chip.x, eth_chip.y):
                        writer.write(f'"{chip.x} {chip.y}",')
            writer.write("\n")

    # create the progress bar for end users
    progress = ProgressBar(n_chips, "Writing clock drift report")

    # iterate over ethernet chips and then the chips on that board
    txrx = FecDataView.get_transceiver()
    with open(directory_name, "a", encoding="utf-8") as writer:
        for eth_chip in eth_chips:
            if ethernet_only:
                __write_drift(txrx, eth_chip, writer)
                progress.update()
            else:
                last_drift = None
                for chip in machine.get_chips_by_ethernet(
                        eth_chip.x, eth_chip.y):
                    drift = __write_drift(txrx, chip, writer)
                    if last_drift is None:
                        last_drift = drift
                    elif last_drift != drift:
                        logger.warning(
                            "On board {}, chip {}, {} is not in sync"
                            " ({} vs {})",
                            eth_chip.ip_address, chip.x, chip.y,
                            drift, last_drift)
                    progress.update()
        writer.write("\n")


def __write_drift(txrx, chip, writer):
    # pylint: disable=protected-access
    drift = txrx._get_sv_data(
        chip.x, chip.y, SystemVariableDefinition.clock_drift)
    drift = struct.unpack("<i", struct.pack("<I", drift))[0]
    drift = drift / (1 << 17)
    writer.write(f'"{drift}",')
    return drift
